# renderers/rtf.py
import struct
from copy import deepcopy

from pygments.token import Token, Whitespace

from purdy.content import Code, Document, RenderState
from purdy.parser import HighlightOn, HighlightOff
from purdy.renderers.formatter import StrFormatter

# ===========================================================================
# RTF Specific Utilities
# ===========================================================================

class RTFFormatter(StrFormatter):
    def __init__(self, rtf_page, section, exceptions):
        self.rtf_page = rtf_page
        exceptions = deepcopy(exceptions)

        hl_colour = section.theme.colour_map[HighlightOn]
        if isinstance(hl_colour, str):
            index, _ = rtf_page.colour_table[hl_colour]
            exceptions.update({
                HighlightOn:  fr"\cf{index} " + "{text}\n",
                HighlightOff: r"\cf0" + "\n",
            })
        else:
            # Close tag using the auto colour (0) for fg & bg, pass thru attrs
            exceptions.update({
                HighlightOn:  self.tag_open(*hl_colour) + "{text}",
                HighlightOff: self.tag_close(hl_colour[2]),
            })

        super().__init__(section, exceptions)
        self.newline = "\\\n"
        self.escape = self.rtf_encode

    @classmethod
    def rtf_encode(cls, text):
        """RTF uses a weird UTF-16 decimal escape sequence for encoding unicode.
        This method take a string and returns an RTF encoded version of it."""
        output = []
        data = text.encode('utf-16le')
        length = len(data) // 2

        # use struct.unpack to turn the pairs of bytes into decimal unicode
        # escape numbers
        parts = struct.unpack(f'<{length}H', data)
        index = 0
        size = len(parts)

        while index < size:
            num = parts[index]

            if num <= 127:
                # regular ascii
                letter = chr(num)

                # Escape characters that are used by RTF
                if letter == '\\':
                    output.append('\\\\')
                elif letter in '{}':
                    output.append('\\' + letter)
                else:
                    output.append(letter)
            elif num <= 256:
                # extended ascii, use hex notation
                output.append(f"\\'{num:2x}" )
            elif num < 55296:
                # 0xD800 (55296) is the the boundary for two-word encoding, less
                # than this means it is a single item
                output.append(f"\\uc0\\u{num}")
            else:
                # greater than 0xD800, means two words
                index += 1
                next_num = parts[index]
                output.append(f"\\uc0\\u{num} \\u{next_num}")

            index += 1

        return ''.join(output)

    def tag_open(self, fg, bg, attrs):
        tag = ""
        if fg:
            index, _ = self.rtf_page.colour_table[fg]
            tag += fr"\cf{index} "
        if bg:
            index, _ = self.rtf_page.colour_table[bg]
            tag += fr"\cb{index} "

        if "bold" in attrs:
            tag += r"\b "

        if "italic" in attrs:
            tag += r"\i "

        return tag

    def tag_close(self, attrs):
        tag = ""
        if "italic" in attrs:
            tag += r"\i0 "

        if "bold" in attrs:
            tag += r"\b0 "

        # Return colouring fg to auto and bg to background
        tag += r"\cf0 \cb1"
        return tag

    def _map_tag(self, token, fg, bg, attrs, exceptions):
        if token in exceptions:
            self.tag_map[token] = exceptions[token]
            return

        # Default handling
        if not (fg or bg or attrs):
            # No formatting
            self.tag_map[token] = "{text}"
            return

        tag = self.tag_open(fg, bg, attrs)
        tag += "{text}"
        tag += self.tag_close(attrs)
        tag += "\n"

        self.tag_map[token] = tag


class RTFPage:
    PREAMBLE = r"{\rtf1\ansi\ansicpg1252" + "\n"
    FONT_TABLE = r"{\fonttbl\f0\fnil\fcharset0 %s;}" + "\n"
    FONT_SPEC = r'\f0\fs28' + '\n'

    def __init__(self, background, container, fontname="RobotoMono-Regular"):
        """RTF uses a look-up table for colouring. This class builds a table
        for lookups and provides mapping utilities."""
        self.has_background = background is not None
        self.fontname = fontname

        self.colour_table = {}
        if self.has_background:
            # Colour 0 is "auto", start counting at 1
            self.colour_table[background] = (1, self.rgb_to_rtf(background))
        else:
            # Default to white
            self.colour_table[background] = (1, self.rgb_to_rtf("ffffff"))

        for section in container:
            for token, value in section.theme.colour_map.items():
                if not value:
                    continue

                if isinstance(value, tuple):
                    # Add the fg and bg colours
                    self._set_colour(value[0])
                    self._set_colour(value[1])
                else:
                    # Single value, add it
                    self._set_colour(value)

    def _set_colour(self, value):
        if not value or value in self.colour_table:
            return

        # Value not in the colour table, create a new one
        pos = len(self.colour_table) + 1
        self.colour_table[value] = (pos, self.rgb_to_rtf(value))

    @classmethod
    def rgb_to_rtf(cls, colour):
        """Translates a three or six digit RGB hex value into the RTF
        equivalent."""
        double_it = lambda colour, index: f"0x{colour[index]}{colour[index]}"

        if len(colour) == 3:
            red = int(double_it(colour, 0), base=16)
            green = int(double_it(colour, 1), base=16)
            blue = int(double_it(colour, 2), base=16)
        else:
            red = int(f"0x{colour[0:2]}", base=16)
            green = int(f"0x{colour[2:4]}", base=16)
            blue = int(f"0x{colour[4:6]}", base=16)

        return fr'\red{red}\green{green}\blue{blue}'

    def header_string(self):
        result = self.PREAMBLE
        result += self.FONT_TABLE % self.fontname

        # Colour table
        result += r"{\colortbl;"

        colour_list = list(self.colour_table.values())
        colour_list.sort()

        for _, rtf_colour in colour_list:
            result += f"{rtf_colour};"

        result += "}\n"

        # Font table
        result += self.FONT_SPEC + "\n"

        # Set background colour (\cb0 is the default colour, \cb1 is the first
        # colour in our table which has been set to the background)
        if self.has_background:
            result += r"\chsdng10000\chcbpat0\cb1" + "\n"

        return result

    def footer_string(self):
        return "}\n"

# ===========================================================================

def to_rtf(container):
    """Transforms tokenized content in a :class:`Code` object into a string
    representation of RTF.

    :param container: `Code` or :class:`MultiCode` object to render
    """
    if isinstance(container, Code):
        container = Document(container)

    render_state = RenderState(container)

    page = RTFPage(container.background, container)
    render_state.content += page.header_string()

    for section in container:
        code_tag_exceptions = {
            Token:      r"\cf0 {text}" + "\n",
            Whitespace: r"\cf0 {text}" + "\n",
        }

        render_state.formatter = RTFFormatter(page, section,
            code_tag_exceptions)
        section.render(render_state)

    render_state.content += page.footer_string()
    return render_state.content



# utils.py
from pathlib import Path

# ===========================================================================
# Cosmetic Code Utils
# ===========================================================================

class TextCodeCleaner:
    @classmethod
    def remove_double_blanks(cls, text, trim_whitespace=True):
        """Removes any lines where two or more are blank in a row.

        :param text: Text to process
        :param trim_whitespace: When True (default) a line with only
            whitespace is considered blank, otherwise it only looks for \\n
        """
        lines = text.split('\n')
        output = []

        previous = 'asdf'
        for line in lines:
            if trim_whitespace \
                    and previous.strip() == '' and line.strip() == '':
                continue
            elif previous == '' and line == '':
                    continue

            output.append(line)
            previous = line

        return '\n'.join(output)

    @classmethod
    def flush_left(cls, text):
        """Removes a consistent amount of leading whitespace from the front of
        each line so that at least one line is flushed left.

        .. warning:: will not work with mixed tabs and spaces
        """
        lines = text.split('\n')
        leads = [len(line) - len(line.lstrip()) for line in lines if \
            len(line.strip())]
        if not leads:
            # only blank lines, do nothing
            return text

        min_lead = min(leads)
        output = []
        for line in lines:
            if len(line.lstrip()):
                output.append(line[min_lead:])
            else:
                output.append(line)

        return '\n'.join(output)


class CodeCleaner:
    @classmethod
    def _get_text(cls, filename):
        if isinstance(filename, Path):
            path = filename
        else:
            path = Path(filename).resolve()

        return path.read_text()

    @classmethod
    def remove_double_blanks(cls, filename, trim_whitespace=True):
        """Opens the given file and removes any lines where two or more are
        blank in a row.

        :param filename: Name of file to process or a :class:`pathlib.Path`
            object. File is read as text.
        :param trim_whitespace: When True (default) a line with only
            whitespace is considered blank, otherwise it only looks for \\n
        """
        text = cls._get_text(filename)
        return TextCodeCleaner.remove_double_blanks(text, trim_whitespace)

    @classmethod
    def flush_left(cls, filename):
        """Open the given file and removes a consistent amount of leading
        whitespace from the front of each line so that at least one line is
        flushed left.

        .. warning:: will not work with mixed tabs and spaces

        :param filename: Name of file to process or a :class:`pathlib.Path`
            object. File is read as text.
        """
        text = cls._get_text(filename)
        return TextCodeCleaner.flush_left(text)



"""
Lexers
======

Additional Lexers not included in Pygments

"""
import re

from pygments.lexer import Lexer
from pygments.lexers import BashSessionLexer
from pygments.token import Generic, Text

# =============================================================================

class DollarBashSessionLexer(BashSessionLexer):
    """Changes behaviour of the BashSessionLexer to assume a prompt uses a
    '$'. This allows better handling of virtual-env style prompts with lots
    of stuff left of the '$'.
    """
    _ps1rgx = re.compile(r'^([^$]*[$]\s)(.*\n?)')


class NewlineLexer(Lexer):
    """Treats each line in the body as a Generic.Output token"""
    name = "Newline output"
    aliases = ["newline"]

    def get_tokens_unprocessed(self, text):
        for token in text.split("\n"):
            yield 0, Generic.Output, token
            yield 0, Text, "\n"



# renderers/plain.py
from purdy.content import Code, Document, RenderState

class PlainFormatter:
    def render_code_line(self, render_state, line):
        result = ""
        for part in line.parts:
            result += part.text

        if line.has_newline:
            result += "\n"

        render_state.content += result


def to_plain(container):
    """Renders content without any formatting.

    :param container: :class:`Document` or :class:`Code` content to be
        rendered

    :returns: rendered string
    """
    if isinstance(container, Code):
        container = Document(container)

    render_state = RenderState(container)
    render_state.formatter = PlainFormatter()
    for section in container:
        section.render(render_state)

    return render_state.content



# arg_helpers.py
#
# Helper utilities for the command line arguments used by 'purdy' and
# 'subpurdy'
from purdy.__init__ import __version__
from purdy.content import Code, Document
from purdy.parser import LexerSpec

# =============================================================================
# Document Factory
# =============================================================================

def document_factory(args, theme_name="default"):
    if args.nocolour:
        # Override theme
        theme_name = "no_colour"

    code = Code(args.filename, args.lexer, theme_name)
    doc = Document(code)

    # Arguments are conditional on the subcommand, so not all args are
    # available all the time
    if hasattr(args, "bg"):
        doc.background = args.bg

    if hasattr(args, "num") and args.num:
        doc.line_numbers_enabled = True
        doc.starting_line_number = args.num

    if hasattr(args, "wrap") and args.wrap:
        doc.wrap = args.wrap

    if hasattr(args, "highlight") and args.highlight:
        code.highlight(*args.highlight)

    return doc

# =============================================================================
# Argument Builders for argparse
# =============================================================================

def lexer_arg(parser):
    lexer_help = (
        "Name of lexer to use to parse the file. Defaults to automaticallly "
        "detecting based on name of file. Choices are:\n"
    )

    for key, value in LexerSpec.built_ins.items():
        lexer_help += f"   * '{key}' -- {value.description}\n"

    for key, value in LexerSpec.aliases.items():
        lexer_help += f"   * '{key}' -- Alias for {value}\n"

    parser.add_argument("-l", "--lexer", choices=LexerSpec.names,
        default="detect", help=lexer_help)


def version_arg(parser):
    parser.add_argument("--version", action="version",
        version="%(prog)s {version}".format(version=__version__ ))


def filename_arg(parser):
    parser.add_argument("filename", help="Name of file to parse")

# -----------------------------------------------------------------------------

def num_arg(parser):
    parser.add_argument("--num", help=("Display line numbers with code "
        "starting with the value given here"), type=int, default=None)


def wrap_arg(parser):
    parser.add_argument("--wrap", help="Wrap line width at this value",
        type=int, default=None)


def no_colour_arg(parser):
    parser.add_argument("--nocolour", "--nc", help=("By default code is "
        "colourized. This flag turns that off."),
        action="store_true")


def highlight_arg(parser):
    parser.add_argument("--highlight", "--hl",
        help=(
            "Highlight a code segment. This argument can be used multiple"
            "times. Each value can be a line number (zero-indexed, negative "
            "indexing supported, a range separated by a dash ('1-3' index 1, "
            "2, and 3 highlighted), or highlight within the line ('3:10,5' "
            "for index 3, highlighting starting at character 10, for 5 "
            "characters"
        ), action="append")


def bg_arg(parser):
    parser.add_argument("--bg", help=(
            "Change the background colour. Indicated using hex RRGGBB format, "
            "like HTML #AABBCC, but without the leading #"
        ), type=str, default=None)

# =============================================================================
# Grouped Argument Builders
# =============================================================================

def general_args(parser):
    lexer_arg(parser)
    version_arg(parser)


def doc_args(parser):
    num_arg(parser)
    wrap_arg(parser)
    no_colour_arg(parser)
    highlight_arg(parser)
    bg_arg(parser)



# themes.py
from copy import copy

from pygments.token import (Keyword, Name, Comment, String, Error, Number,
    Operator, Generic, Token, Whitespace, Punctuation, Text, Literal)

from purdy.parser import HighlightOn, HighlightOff, Fold, LineNumber

# ===========================================================================
# Theme Class
# ===========================================================================

class Theme:
    def __init__(self, full_name, colour_map, inherit=None):
        self.full_name = full_name

        if inherit is not None:
            # Inherit from another theme
            self.colour_map = copy(inherit.colour_map)
        else:
            self.colour_map = {}

        self.colour_map.update(colour_map)

    def values(self):
        for key, value in self.colour_map.items():
            if isinstance(value, tuple):
                fg, bg, attrs = value
            else:
                fg = value
                bg = ""
                attrs = ""

            yield key, fg, bg, attrs

# ===========================================================================
# Empty Theme
# ===========================================================================

EMPTY_THEME = Theme("empty", {})
no_colour = Theme("no_colour", {})

# ===========================================================================
# Default Themes
# ===========================================================================

default_code = Theme("default_code", {
    Token:               "",
    Whitespace:          "",
    Comment:             "66dddd",
    Keyword:             "dd88dd",
    Keyword.Constant:    "008000",
    Operator:            "aaaaaa",
    Punctuation:         "88ddff",
    Text:                "dddddd",
    Name:                "dddddd",
    Name.Builtin:        "88aaff",
    Name.Builtin.Pseudo: ("aa6666", "", "bold"),
    Name.Function:       "aaddff",
    Name.Class:          "aaddff",
    Name.Exception:      ("ffdd66", "", "bold"),
    Name.Decorator:      ("ffdd66", "", "bold"),
    String:              "dddddd",
    Number:              "ff8866",
    Generic.Prompt:      ("ffffff", "", "bold"),
    Generic.Error:       ("ffdd66", "", "bold"),
    Generic.Traceback:   "dddddd",
    Error:               ("ffdd66", "", "bold"),

    # Purdy tokens
    HighlightOn:         ("", "ffffff", ""),
    HighlightOff:        "",
    Fold:                "ffffff",
    LineNumber:          "7f7f7f",
})


default_doc = Theme("default_doc", {
    Name.Tag:           "cdcd00",
    Name.Attribute:     "cdcd00",
    Literal:            "88aaff",
    Generic.Heading:    "cdcd00",
    Generic.Subheading: "cdcd00",
    Generic.Emph:       "000087",
    Generic.Strong:     "005f00",
    String:             "008000",
}, default_code)


default_xml = Theme("default_xml", {
    Name.Attribute: "cdcd00",
    Keyword:        "88aaff",
    Name.Tag:       "88aaff",
    Punctuation:    "88aaff",
}, default_code)

# ===========================================================================
# RTF Themes
# ===========================================================================

rtf_code = Theme("rtf_code", {
    Token:              "",
    Whitespace:         "",
    Comment:            ("8f5902", "", "italic"),
    Keyword:            ("2048a7", "", "bold"),
    Operator:           ("ce5c00", "", "bold"),
    Name:               "000000",
    Name.Builtin:       "3465a4",
    Name.Function:      "000000",
    Name.Class:         "000000",
    Name.Exception:     ("cc0000", "", "bold"),
    Name.Decorator:     ("5c35cc", "", "bold"),
    Name.Variable:      "",
    Name.Constant:      "",
    Name.Attribute:     "",
    Name.Tag:           "",
    Punctuation:        ("000000", "", "bold"),
    String:             "4e9a06",
    String.Doc:         ("8f5902", "", "italic"),
    Number:             ("0000cf", "", "bold"),
    Generic.Prompt:     "8f5902",
    Generic.Error:      "ef2929",
    Generic.Traceback:  ("a40000", "", "bold"),
    Error:              "ef2929",

    # Purdy tokens
    HighlightOn:         ("", "d5d5d5", ""),
    HighlightOff:        "",
    Fold:                "000000",
    LineNumber:          "000000",
})

rtf_xml = Theme("rtf_xml", {
    Name.Attribute: ("2048a7", "", "bold"),
    Name.Tag:       "3465a4",
    Keyword:        "3465a4",
    Punctuation:    "3465a4",
}, rtf_code)

rtf_doc = Theme("rtf_doc", {
    Generic.Emph:       "ef2929",
    Generic.Strong:     ("a40000", "", "bold"),
    Literal:            "",
    Generic.Heading:    "a40000",
    Generic.Subheading: "a40000",
}, rtf_code)

# ===========================================================================
# Python REPL Theme (as of 3.14)
# ===========================================================================

pyrepl_code = Theme("pyrepl_code", {
    Token:               "",
    Whitespace:          "",

    Comment:             "c91b00",
    Keyword:             ("6972ff", "", "bold"),
    Name.Builtin:        "00c5c8",
    Name.Builtin.Pseudo: "c7c7c7",
    Name.Class:          ("ffffff", "", "bold"),
    Name.Decorator:      "00c5c8",
    Name.Exception:      "00c5c8",
    Name.Function:       ("ffffff", "", "bold"),
    Number:              "c7c400",
    Punctuation.Marker:  ("ff6e67", "", "bold"),
    String:              "00c200",

    Operator:            "c7c7c7",
    Operator.Word:       ("6972ff", "", "bold"),
    Generic.Prompt:      ("ff77ff", "", "bold"),
    Generic.Error:       ("ff77ff", "", "bold"),
    Error:               ("cc0000", "", "bold"),
})

# ===========================================================================
# Theme Map
# ===========================================================================

THEME_MAP = {
    "default": {
        "code": default_code,
        "doc": default_doc,
        "xml": default_xml,
    },
    "no_colour": {
        "code": no_colour,
        "doc": no_colour,
        "xml": no_colour,
    },
    "no_color": {
        "code": no_colour,
        "doc": no_colour,
        "xml": no_colour,
    },
    "rtf": {
        "code": rtf_code,
        "doc": rtf_doc,
        "xml": rtf_xml,
    },
    "pyrepl": {
        "code": pyrepl_code,
    },
}



from unittest import TestCase

from pygments.token import Keyword, Comment

from purdy.themes import Theme

# =============================================================================

class TestThemes(TestCase):
    def test_themes(self):

        # Theme inheritance
        theme = Theme("first", {Comment: "778899"})
        second = Theme("second", {Keyword: "889900"}, theme)

        self.assertEqual(2, len(second.colour_map))
        self.assertIn(Comment, second.colour_map)
        self.assertIn(Keyword, second.colour_map)

        # Values iterator
        theme = Theme("third", {
            Comment: "112233",
            Keyword: ("223344", "556677", "bold"),
        })

        expected = [
            (Comment, "112233", "", ""),
            (Keyword, "223344", "556677", "bold"),
        ]
        result = list(theme.values())
        self.assertEqual(expected, result)



from purdy.content import Code
from purdy.tui.apps import AppFactory
from purdy.tui.codebox import BoxSpec, EscapeText, RowSpec, TextSection

__all__ = [
    "AppFactory", "BoxSpec", "Code", "EscapeText", "RowSpec", "TextSection"
]



# purdy.tui.widgets.py
import math

from rich.color import Color
from rich.segment import Segment, Segments
from rich.style import Style

from textual.app import ComposeResult
from textual.containers import Grid, ScrollableContainer, Container
from textual.scrollbar import ScrollBarRender
from textual.widgets import Static

# =============================================================================
# Scroll Renderers
# =============================================================================

class DarkThinBarRender(ScrollBarRender):
    @classmethod
    def render_bar(cls, size, virtual_size, window_size, position,
            thickness, vertical, back_color, bar_color) -> Segments:
        # Hard code the thickness and colour for now
        return super().render_bar(size, virtual_size, window_size,
            position, 1, vertical, back_color, Color.parse("#333333"))


class LightThinBarRender(ScrollBarRender):
    @classmethod
    def render_bar(cls, size, virtual_size, window_size, position,
            thickness, vertical, back_color, bar_color) -> Segments:
        # Hard code the thickness and colour for now
        return super().render_bar(size, virtual_size, window_size,
            position, 1, vertical, back_color, Color.parse("#999999"))


class TriangleScrollRender(ScrollBarRender):
    NO_INDICATOR = "▮"
    UP_INDICATOR = "▲"
    DOWN_INDICATOR = "▼"
    COLOUR = Color.parse("white")
    BLANK = " "
    LINE_INDICATORS = ["⎺", "⎻", "⎼", "⎽"]

    @classmethod
    def render_bar(
        cls,
        size: int = 25,
        virtual_size: float = 50,
        window_size: float = 20,
        position: float = 0,
        thickness: int = 1,
        vertical: bool = True,
        back_color: Color = Color.parse("#555555"),
        bar_color: Color = Color.parse("bright_magenta"),
    ) -> Segments:

        if not (window_size and size and virtual_size and size != virtual_size):
            style = Style(bgcolor=back_color)
            segments = [Segment(cls.BLANK, style=style)] * int(size)
            segments[0] = Segment(cls.NO_INDICATOR, Style(color=cls.COLOUR))
            return Segments(segments, new_lines=True)

        bg_segment = Segment(cls.BLANK, Style(bgcolor=back_color))
        segments = [bg_segment] * int(size)
        segments[0] = Segment(cls.UP_INDICATOR, Style(color=cls.COLOUR))
        segments[-1] = Segment(cls.DOWN_INDICATOR, Style(color=cls.COLOUR))

        # Calculate where the position indicator will go, ensuring it is only
        # on the top or bottom when fully scrolled up or down
        position_ratio = position / (virtual_size - window_size)
        index = int(size * position_ratio)
        if position == 0:
            # Full top position
            segments[0] = Segment(cls.UP_INDICATOR,
                Style(color=cls.COLOUR, underline=True))
        elif index == size:
            # Full bottom position (index calc overreaches)
            segments[-1] = Segment(cls.DOWN_INDICATOR,
                Style(color=cls.COLOUR, underline=True))
        else:
            # Somewhere in-between, position the indicator in the 2nd through
            # 2nd-last positions
            scroll_height = size - 2

            # Number of lines each tick on the scroll bar represents
            line_size = (virtual_size - window_size) / scroll_height
            slot = math.ceil(position / line_size)

            # Fractional position within a line, used to map to the
            # LINE_INCREMENTS so that it looks like it is moving even if it is
            # in the same slot
            frac_line_pos = (position % line_size) / line_size
            if frac_line_pos == 0:
                # Perfect division gives 0, but that is the bottom not the top
                frac_line_pos = 1

            # Convert to an index in the inter-line indicators by apply it as
            # a fraction to the number of indicators, answer will be 1-indexed
            # unless frac_line_pos is zero, so compensate
            pos_index = math.ceil(frac_line_pos * len(cls.LINE_INDICATORS))
            if pos_index != 0:
                pos_index -= 1

            # Slot is 1-indexed, but we need to start in the second position
            # anyway
            segments[slot] = Segment(cls.LINE_INDICATORS[pos_index],
                Style(color=cls.COLOUR))

        return Segments(segments, new_lines=True)


class BlurredTriangleScrollRender(TriangleScrollRender):
    NO_INDICATOR = "▯"
    UP_INDICATOR = "△"
    DOWN_INDICATOR = "▽"
    COLOUR = Color.parse("#555555")


class AlwaysVerticalScroll(ScrollableContainer):
    """A container with vertical layout and an always-on scrollbar on the Y
    axis."""

    DEFAULT_CSS = """
    AlwaysVerticalScroll {
        layout: vertical;
        overflow-x: hidden;
        overflow-y: scroll;
    }
    """

    def on_focus(self, event):
        self.vertical_scrollbar.renderer = TriangleScrollRender

    def on_blur(self, event):
        self.vertical_scrollbar.renderer = BlurredTriangleScrollRender

# =============================================================================
# Purdy Container
#
# Single container for all the CodeWidgets with a transition overlay

class PurdyContainer(Container):
    def __init__(self, owner, row_specs, max_height=None):
        # Import here to avoid circular import
        from purdy.tui.codebox import CodeBox

        super().__init__()
        self.owner = owner
        self.row_specs = row_specs
        self.max_height = max_height

        self.grid_width = 0
        self.grid_height = 0

        for row_num, row_spec in enumerate(self.row_specs):
            self.grid_height += row_spec.height

            self.owner.rows.append([])
            for box_num, box_spec in enumerate(row_spec.boxes):
                if row_num == 0:
                    # Use the first row as the blueprint for the width
                    self.grid_width += box_spec.width

                box = CodeBox(f"r{row_num}_b{box_num}", row_spec, box_spec)
                self.owner.rows[-1].append(box)

    def compose(self) -> ComposeResult:
        if self.max_height is not None:
            self.styles.max_height = self.max_height
            print(self.styles)

        with Container(classes="pc_inner") as self.inner:
            self.overlay = Container(classes="pc_effect_overlay")
            yield self.overlay

            with Grid(classes="pc_grid") as self.grid:
                # Loop through the CodeBoxes and compose their widgets
                for row in self.owner.rows:
                    for box in row:
                        yield box.widget

                self.grid.styles.grid_size_rows = self.grid_height
                self.grid.styles.grid_size_columns = self.grid_width

# =============================================================================
# Code Widget
#
# Scrollable container that displays one :class:`Document` and has a
# transition overlay

class CodeWidget(Container):
    def __init__(self, border="", title=None):
        super().__init__(classes="code_widget")

        self.border = border.lower()
        self.pad_top = 0
        self.pad_bottom = 0
        self.pad_left = 0
        self.pad_right = 0

        self.title = title

        self.code_display = Static("", classes="code_display")

    def compose(self) -> ComposeResult:
        with Container(classes="cw_container") as self.container:
            self.overlay = Container(classes="cw_effect_overlay")
            yield self.overlay

            with Container(classes="code_holder") as self.code_holder:
                print("title is", self.title)
                if self.title is not None:
                    yield Static(self.title, classes="code_title")
                    print("yielded widget")

                with AlwaysVerticalScroll(classes="vscroller") as self.vs:
                    self.vs.vertical_scrollbar.renderer = \
                        BlurredTriangleScrollRender

                    yield self.code_display

                    if "t" in self.border:
                        self.vs.styles.border_top = ("solid", "white")
                        self.pad_top = 1
                    if "b" in self.border:
                        self.vs.styles.border_bottom = ("solid", "white")
                        self.pad_bottom = 1
                    if "r" in self.border:
                        self.vs.styles.border_right = ("solid", "white")
                        self.pad_right = 1
                    if "l" in self.border:
                        self.vs.styles.border_left = ("solid", "white")
                        self.pad_left = 1

    async def on_key(self, event):
        key = event.key
        match key:
            case "alt+pagedown":
                # Scroll down half a page
                pos = (self.vs.scroll_y +
                    self.vs.scrollable_content_region.height // 2)
                self.vs.scroll_to(y=pos)

                # Eat event
                event.stop()
            case "alt+pageup":
                # Scroll up half a page
                pos = (self.vs.scroll_y -
                    self.vs.scrollable_content_region.height // 2)
                self.vs.scroll_to(y=pos)

                # Eat event
                event.stop()



from pathlib import Path
from unittest import TestCase

from purdy.utils import CodeCleaner, TextCodeCleaner

# ===========================================================================
# Result Constants
# ===========================================================================

WITHOUT_DOUBLES = """\
one

two

three

four

five below whitespace line
"""

WITHOUT_DOUBLES_WHITESPACE = """\
one

two

three

four

    
five below whitespace line
"""

FLUSHED = """\
one
    two
        three
    four
five
"""

# =============================================================================

class TestCleaners(TestCase):
    # Test CodeCleaner which calls TextCodeCleaner for everything

    def test_remove_double(self):
        path = (Path(__file__).parent / Path("data/doubles.txt")).resolve()

        # Whitespace is considered a blank line, testing with a Path object
        result = CodeCleaner.remove_double_blanks(path)
        self.assertEqual(WITHOUT_DOUBLES, result)

        # Do it again but with the filename instead of a Path
        result = CodeCleaner.remove_double_blanks(str(path))
        self.assertEqual(WITHOUT_DOUBLES, result)

        # Whitespace NOT a blank line, testing with a Path object
        result = CodeCleaner.remove_double_blanks(path, False)
        self.assertEqual(WITHOUT_DOUBLES_WHITESPACE, result)

        # Do it again but with the filename instead of a Path
        result = CodeCleaner.remove_double_blanks(str(path), False)
        self.assertEqual(WITHOUT_DOUBLES_WHITESPACE, result)

    def test_flush_left(self):
        path = (Path(__file__).parent / Path("data/flush.txt")).resolve()

        # Test with Path object
        result = CodeCleaner.flush_left(path)
        self.assertEqual(FLUSHED, result)

        # Do it again but with the filename instead of a Path
        result = CodeCleaner.flush_left(str(path))
        self.assertEqual(FLUSHED, result)

        # No identing at all
        expected = "one\ntwo\nthree\n"
        result = TextCodeCleaner.flush_left(expected)
        self.assertEqual(expected, result)

        # No flushing needed
        expected = "one\n   two\n   three\n"
        result = TextCodeCleaner.flush_left(expected)
        self.assertEqual(expected, result)

        # Empty string doesn't blow up
        result = TextCodeCleaner.flush_left("")
        self.assertEqual("", result)



# purdy.tui.tuicontent.py
from textual.content import Content as TContent

from purdy.content import Section
from purdy.parser import LineNumber
from purdy.themes import THEME_MAP

# =============================================================================
# Content Abstractions
# =============================================================================

class EscapeText(str):
    """Class to distinguish plain text from Textual markup to present escaped
    text."""
    pass


class TextSection(Section):
    """Abstracts one or more lines of Textual markup strings"""
    default_theme_name = "default"
    default_theme_category = "code"

    def __init__(self, content=None, theme=None):
        super().__init__()

        # Set the content
        if content is None:
            self.lines = []
        elif isinstance(content, (str, EscapeText)):
            self.lines = [content, ]
        elif isinstance(content, list):
            self.lines = content
        else:
            raise ValueError("TextSection requires string or list of strings")

        # Configure Theme
        if theme is None:
            theme = self.default_theme_name

        if isinstance(theme, str):
            self.theme = THEME_MAP[theme][self.default_theme_category]
        else:
            self.theme = theme

    def render_line(self, render_state, line, line_index):
        if render_state.doc.line_numbers_enabled:
            num = render_state.next_line_number()
            render_state.content += render_state.formatter.part_to_content(
                LineNumber, num)

        if isinstance(line, EscapeText):
            render_state.content += TContent.from_markup("$text", text=line)
        else:
            render_state.content += TContent.from_markup(line)

        render_state.content += "\n"



from pathlib import Path
from unittest import TestCase

import shared

# =============================================================================

class TestRenderers(TestCase):
    def test_renderers(self):
        compare_dir = Path(__file__).parent / Path("compare")

        for name in shared.RENDER_TESTS:
            # Read the comparison file
            path = (compare_dir / Path(f"{name}.out")).resolve()
            expected = path.read_text()

            # Generate the output to compare against
            fn = getattr(shared, f"generate_{name}")
            text = fn()

            try:
                # Convert to list as it makes finding the problem easier
                nexpected = expected.split("\n")
                ntext = text.split("\n")
                self.assertEqual(nexpected, ntext)
            except AssertionError: # pragma: no cover
                print(f"*** Failed when testing {name}")
                raise



from unittest import TestCase

from pygments.token import Comment, Keyword, Name

from purdy.content import Code, Document
from purdy.renderers.rtf import RTFPage, RTFFormatter
from purdy.themes import Theme

# =============================================================================

class TestRTF(TestCase):
    def test_rtf_colour(self):
        theme = Theme("dummy_theme", {
            Comment: "112233",
            Keyword: ("223344", "556677", "bold"),
            Name: "abc",
        })

        code = Code.text("", "plain")
        code.theme = theme
        doc = Document(code)

        page = RTFPage("222222", doc)

        # Test map (0 is the auto value, starts at 1)
        expected = {
            "222222": (1, r"\red34\green34\blue34"),
            "112233": (2, r"\red17\green34\blue51"),
            "223344": (3, r"\red34\green51\blue68"),
            "556677": (4, r"\red85\green102\blue119"),
            "abc": (5, r"\red170\green187\blue204"),
        }
        self.assertEqual(expected, page.colour_table)

        # Test table
        expected = (
            r"{\colortbl;"
            r"\red34\green34\blue34;"
            r"\red17\green34\blue51;"
            r"\red34\green51\blue68;"
            r"\red85\green102\blue119;"
            r"\red170\green187\blue204;"
            r"}"
        )
        self.assertIn(expected, page.header_string())

    def test_encode(self):
        # ASCII text
        self.assertEqual("a", RTFFormatter.rtf_encode("a"))
        self.assertEqual(r"a\\b", RTFFormatter.rtf_encode(r"a\b"))

        # ASCII above 127
        c = chr(200)  # hex C8
        self.assertEqual(r"\'c8", RTFFormatter.rtf_encode(c))

        # Unicode above 256 but below word boundary
        c = chr(257)
        self.assertEqual(r"\uc0\u257", RTFFormatter.rtf_encode(c))

        # Unicode above 0xD800 is two words
        c = "😂"
        self.assertEqual(r"\uc0\u55357 \u56834", RTFFormatter.rtf_encode(c))



# purdy.tui.apps.py
from textual.app import App, ComposeResult

from purdy.tui.animate import AnimationController, cell_list
from purdy.tui.codebox import BoxSpec, RowSpec
from purdy.tui.purdybox import PurdyBox

# =============================================================================

class PurdyApp(App):
    CSS_PATH = "style.tcss"

    def __init__(self, row_specs, max_height=None):
        super().__init__()
        self.row_specs = row_specs

        # Build the PurdyBox containing all the CodeBoxes first so that
        # they can be populated before Textual mounts everything, then yield
        # it in `compose`
        self.control = PurdyBox(self.row_specs, max_height)

    def compose(self) -> ComposeResult:
        yield self.control.container

    async def on_mount(self):
        # Force focus to our first CodeBox, then start animation
        self.set_focus(self.control.rows[0][0].widget.code_display)
        await self.controller.forwards()

    def run(self):
        self.controller = AnimationController(self)
        super().run()

    def _debug_info(self):
        output = ["==== DEBUG INFO ===="]
        from purdy.content import Code

        for row_index, row in enumerate(self.control.rows):
            for box_index, box in enumerate(row):
                output.append(f"\nRow {row_index} Box {box_index}")
                for section in box.doc:
                    if not isinstance(section, Code):
                        output.append(f"   {section.__class__.__name__}")
                        for line in section.lines:
                            output.append((f"      *{repr(line)}*"))
                    else:
                        output.append("   Code")
                        for line in section.lines:
                            output.append("      Line")
                            for part in line.parts:
                                output.append((
                                    f"      {part.token} : {part.text}*"
                                ))

        output.append("\nCell list:")
        for count, cell in enumerate(cell_list):
            output.append(f"   {count} {cell}")

        return "\n".join(output)

    async def on_key(self, event):
        key = event.key
        match key:
            case "q" | "Q":
                exit()
            case "right":
                await self.controller.forwards()
            case "left":
                await self.controller.backwards()
            case "s":
                await self.controller.skip()
            case "d":
                if "devtools" in self.features:
                    print(self._debug_info())

# =============================================================================
# Factory Methods
# =============================================================================

class AppFactory:
    @classmethod
    def simple(cls, max_height=None, line_number=None, auto_scroll=False,
            title=None):
        """Creates a purdy display with a single code box inside of it.

        :param max_height: limit the height of the content in the terminal,
            defaults to None
        :param line_number: specify a starting line number for the code box,
            defaults to None
        :param auto_scroll: when True, if content is added to the code box it
            scrolls to the bottom. Defaults to False.
        :param title: a title to include at the top of the codebox, defaults
            to None
        """
        row_specs = [
            RowSpec(1, [BoxSpec(1, line_number, auto_scroll, title=title)])
        ]

        app = PurdyApp(row_specs, max_height)
        app.box = app.control.rows[0][0]
        return app

    @classmethod
    def split(cls, max_height=None, line_number_top=None, auto_scroll_top=False,
            relative_height_top=1, line_number_bottom=None,
            auto_scroll_bottom=False, relative_height_bottom=1,
            top_title=None, bottom_title=None):
        """Creates a purdy display with two code boxes one on top of the other.

        :param max_height: limit the height of the content in the terminal,
            defaults to None
        :param line_number_top: specify a starting line number for the top code
            box, defaults to None
        :param auto_scroll_top: when True, if content is added to the top code
            box it scrolls to the bottom. Defaults to False.
        :param relative_height_top: specify the height of the top code box
            against the bottom one. For example top=1, bottom=1 makes them the
            same size, top=3, bottom=1, makes them be 3/4 and 1/4 of the screen
            respectively
        :param line_number_bottom: specify a starting line number for the bottom
            code box, defaults to None
        :param auto_scroll_bottom: when True, if content is added to the bottom
            code box it scrolls to the bottom. Defaults to False.
        :param relative_height_bottom: specify the height of the bottom code box
            against the top one.
        :param top_title: a title to include over the top codebox, defaults
            to None
        :param bottom_title: a title to include over the bottom codebox,
            defaults to None
        """
        row_specs = [
            RowSpec(relative_height_top, [
                BoxSpec(1, line_number_top, auto_scroll_top, "b",
                    title=top_title)]),
            RowSpec(relative_height_bottom, [
                BoxSpec(1, line_number_bottom, auto_scroll_bottom,
                    title=bottom_title)])
        ]

        app = PurdyApp(row_specs, max_height)
        app.top = app.control.rows[0][0]
        app.bottom = app.control.rows[1][0]
        return app

    @classmethod
    def full(cls, row_specs, max_height=None):
        """Creates a TUI screen based on a list of rows, where each row is a
        list of :class:`BoxSpec` classes.

        :param row_specs: list of :class:`RowSpec` objects which describe the
            layout information for each row in the resulting display grid. Each
            `RowSpec` contains a list of :class:`BoxSpec` objects, the sum of
            their widths in each row must be equal.

        :param max_height: Max height of the app within your terminal.
            Defaults to None.
        """
        app = PurdyApp(row_specs, max_height)
        app.rows = app.control.rows
        return app



# typewriter.py
#
# Utility class for emulating the typing of a Code object
from copy import copy, deepcopy
from collections import namedtuple

from pygments.token import Comment, Generic, Text, Whitespace
from textual.content import Content as TContent
from textual.markup import MarkupTokenizer

from purdy.parser import CodePart, CURSOR, CURSOR_CHAR, LineNumber, token_is_a
from purdy.renderers.textual import TextualFormatter, _CODE_TAG_EXCEPTIONS
from purdy.tui.tui_content import EscapeText

# ===========================================================================

TypewriterOutput = namedtuple("TypewriterOutput", ["text", "state"])


class _CodeTypewriter:
    def __init__(self, base_render_state, src_code, skip_comments,
            skip_whitespace):
        self.base_render_state = copy(base_render_state)
        self.base_render_state.formatter = TextualFormatter(src_code,
            _CODE_TAG_EXCEPTIONS)

        self.src_code = src_code
        self.skip_comments = skip_comments
        self.skip_whitespace = skip_whitespace

        self.is_console = src_code.parser.lexer_spec.console

    def _skip_part(self, part, state="P"):
        self.typing_line.parts.append(part)
        self.typing_line.parts.append(CURSOR)

        self.typing_rs.formatter.render_code_line(self.typing_rs,
            self.typing_line)

        result = TypewriterOutput(
            text=self.cached_rs.content + self.typing_rs.content,
            state=state
        )
        self.results.append(result)

        # Remove the cursor, then reset for # the next line
        del self.typing_line.parts[-1]  # cursor
        self.typing_rs.content = TContent()

    def _run(self):
        """Returns a list of :class:`textual.content.Content` objects
        representing a series of steps used to emulate typing of this source
        object.  """
        self.results = []
        self.cached_rs = copy(self.base_render_state)
        self.cached_rs.content = TContent()

        for src_line in self.src_code.lines:
            self.typing_rs = copy(self.cached_rs)
            self.typing_rs.content = TContent()
            self.typing_line = src_line.spawn()

            if self.typing_rs.doc.line_numbers_enabled:
                num = self.cached_rs.next_line_number()
                self.cached_rs.content += \
                    self.typing_rs.formatter.part_to_content(LineNumber, num)

            # Handle multi-line console output
            if self.is_console:
                first_token = src_line.parts[0].token

                is_output = token_is_a(first_token, Generic.Output)
                is_output |= not token_is_a(first_token, Generic.Prompt)

                if is_output:
                    self.typing_rs.formatter.render_code_line(self.typing_rs,
                        src_line)

                    result = TypewriterOutput(
                        text=self.cached_rs.content + self.typing_rs.content,
                        state=None
                    )
                    self.results.append(result)

                    self.cached_rs.formatter.render_code_line(self.cached_rs,
                        src_line)
                    self.typing_rs.content = TContent()
                    continue

            # Typewriter-ize the line's parts
            for src_part in src_line.parts:
                src_token = src_part.token
                part = CodePart(token=src_token, text="")

                if self.is_console and token_is_a(src_token, Generic.Prompt):
                    # Don't animate prompts
                    self._skip_part(src_part, state="W")
                    continue

                if self.skip_comments and token_is_a(src_token, Comment):
                    # Don't animate comments
                    self._skip_part(src_part)
                    continue

                if self.skip_whitespace:
                    # Don't animate whitespace; this can be a specific token,
                    # or just blank text
                    skip_it = token_is_a(src_token, Whitespace)

                    if token_is_a(src_token, Text):
                        skip_it |= src_part.text.isspace()

                    if skip_it:
                        self._skip_part(src_part)
                        continue

                for char in src_part.text:
                    # Add the character to the part, put it in the line and
                    # put that in the results listing
                    part.text += char
                    self.typing_line.parts.append(part)
                    self.typing_line.parts.append(CURSOR)

                    self.typing_rs.formatter.render_code_line(self.typing_rs,
                        self.typing_line)

                    result = TypewriterOutput(
                        text=self.cached_rs.content + self.typing_rs.content,
                        state="P"
                    )
                    self.results.append(result)

                    # Remove the partial part and the cursor, then reset for
                    # the next line
                    del self.typing_line.parts[-1]  # cursor
                    del self.typing_line.parts[-1]  # partial part
                    self.typing_rs.content = TContent()

                self.typing_line.parts.append(src_part)

            # Update the cached result with the final value of line
            self.cached_rs.formatter.render_code_line(self.cached_rs,
                self.typing_line)

        return self.results

    def _old_run(self):
        """Returns a list of :class:`textual.content.Content` objects
        representing a series of steps used to emulate typing of this source
        object.  """
        prev_was_output = False
        for src_line in self.src_code.lines:
            # Handle multi-line console output
            if self.is_console:
                first_token = src_line.parts[0].token

                is_output = token_is_a(first_token, Generic.Output)
                is_output |= not token_is_a(first_token, Generic.Prompt)

                if is_output:
                    if prev_was_output:
                        # nth output line, don't commit a new Code object,
                        # just update the previous one
                        prev_code = self.results[-1]
                        line = deepcopy(src_line)
                        prev_code.lines.append(line)
                        continue
                    else:
                        # First output line
                        prev_was_output = True
                        self.current.lines.append(src_line)
                        self.results.append(self.current)

                        # Reset for the next pass
                        self.current = deepcopy(self.current)
                        continue
                else:
                    prev_was_output = False

            self.line = src_line.spawn()

            for src_part in src_line.parts:
                src_token = src_part.token
                self.part = CodePart(token=src_token, text="")

                if self.is_console and token_is_a(src_token, Generic.Prompt):
                    # Don't animate prompts
                    self._skip_part(src_part.text)
                    continue

                if self.skip_comments and token_is_a(src_token, Comment):
                    # Don't animate comments
                    self._skip_part(src_part.text)
                    continue

                if self.skip_whitespace:
                    # Don't animate whitespace; this can be a specific token,
                    # or just blank text
                    skip_it = token_is_a(src_token, Whitespace)

                    if token_is_a(src_token, Text):
                        skip_it |= src_part.text.isspace()

                    if skip_it:
                        self._skip_part(src_part.text)
                        continue

                for char in src_part.text:
                    # Add the character to the part, put it in the line and
                    # put that in the results listing
                    self.part.text += char
                    self._commit_part()

                    del self.current.lines[-1]
                    del self.line.parts[-1]

                self.line.parts.append(self.part)
            self.current.lines.append(self.line)

        return self.results


def code_typewriterize(render_state, src_code, skip_comments=True,
        skip_whitespace=True):
    """Outputs a list of :class:`textual.content.Content` objects to be used
    as typing animations"""
    tw = _CodeTypewriter(render_state, src_code, skip_comments, skip_whitespace)
    return tw._run()

# ---------------------------------------------------------------------------

MARKUP_CURSOR = "[white]" + CURSOR_CHAR + "[/]"

def _plain_typewriterize(results, base_render_state, rendered, line):
    current = TContent()

    if base_render_state.doc.line_numbers_enabled:
        num = base_render_state.next_line_number()
        current += base_render_state.formatter.part_to_content(LineNumber, num)

    for char in line:
        current += char
        results.append(rendered + current + CURSOR_CHAR)

    return current


def textual_typewriterize(base_render_state, section):
    results = []
    rendered = TContent()
    tokenizer = MarkupTokenizer()
    tag = ""

    for line in section.lines:
        if isinstance(line, EscapeText):
            # Handle plain text as a special case
            result = _plain_typewriterize(results, base_render_state, rendered,
                line)
            rendered += result
            continue

        # Non-escaped text, possibility that it contains markup
        current = ""

        if base_render_state.doc.line_numbers_enabled:
            num = base_render_state.next_line_number()
            rendered += base_render_state.formatter.part_to_content(LineNumber,
                num)

        for token in tokenizer(line, ("inline", "")):
            if token.name == "text":
                for char in token.value:
                    current += char
                    output = TContent.from_markup(current + MARKUP_CURSOR)
                    results.append(rendered + output)
            elif token.name == "eof":
                pass
            elif token.name == "end_tag":
                tag += token.value
                current += tag
                output = TContent.from_markup(current + MARKUP_CURSOR)
                results.append(rendered + output)
                tag = ""
            else:
                # Token is some part of a tag, accumulate it in current
                tag += token.value

        rendered += TContent.from_markup(line + "\n")

    return results



from unittest import TestCase

from pygments.token import Comment, Keyword, Name, Whitespace

from purdy.content import Code, Document, RenderState
from purdy.parser import CodePart, Parser
from purdy.renderers.formatter import StrFormatter
from purdy.themes import Theme

from shared import code_liner

# =============================================================================

class DummyFormatter(StrFormatter):
    def _map_tag(self, token, fg, bg, attrs, exceptions):
        if token in exceptions:
            self.tag_map[token] = exceptions[token]
            return

        self.tag_map[token] = "text='{text}' " + \
            f"fg:'{fg}' bg:'{bg}' attrs:'{attrs}'"

DUMMY_FORMATTED = """\
not_ancestor
text='comment' fg:'112233' bg:'' attrs:''
text='keyword' fg:'223344' bg:'556677' attrs:'bold'
no_percent
"""

DUMMY_FORMATTED_ESCAPED = """\
<not_ancestor>
text='<comment>' fg:'112233' bg:'' attrs:''
text='<keyword>' fg:'223344' bg:'556677' attrs:'bold'
no_percent
"""

# =============================================================================

class TestFormatter(TestCase):
    def test_formatter(self):
        parser = Parser("py")
        lexer_spec = parser.lexer_spec

        theme = Theme("dummy", {
            Comment: "112233",
            Keyword: ("223344", "556677", "bold"),
            Name: "667788",
        })

        # Hack up a Code object to use our theme and our hand crafted lines
        code = Code.text("", "plain")
        code.theme = theme
        code.lines = [
            code_liner(lexer_spec, True, CodePart(Whitespace, 'not_ancestor')),
            code_liner(lexer_spec, True, CodePart(Comment, 'comment')),
            code_liner(lexer_spec, True, CodePart(Keyword, 'keyword')),
            code_liner(lexer_spec, True, CodePart(Name, 'name')),
        ]
        doc = Document(code)
        section = doc[0]

        exceptions = {
            Name: "no_percent",
        }

        # Default theme
        rs = RenderState(doc)
        rs.formatter = DummyFormatter(section, exceptions)
        section.render(rs)
        self.assertEqual(DUMMY_FORMATTED, rs.content)

        # Test with escape
        rs = RenderState(doc)
        rs.formatter = DummyFormatter(section, exceptions)
        rs.formatter.escape = lambda x: f"<{x}>"
        section.render(rs)
        self.assertEqual(DUMMY_FORMATTED_ESCAPED, rs.content)



# _debug.py
# Debug Utilities
from rich.console import Console

# =============================================================================

console = Console(highlight=False)
rprint = console.print

# =============================================================================
# Print Debug Info
# =============================================================================

def print_cr(text, title=""):
    output = text.replace("\n", "↩️\n")
    console.rule(title)
    rprint(output, style="reverse")
    console.rule()


def compare_cr(first, second, title=""):
    console.rule(title)
    first = first.replace("\n", "↩️\n")
    rprint(first, style="reverse")
    console.rule(style="blue")

    second = second.replace("\n", "↩️\n")
    rprint(second, style="reverse")
    console.rule()


def print_plain_output(code, title=""):
    # Prints output from the plain.TextCode
    console.rule(title)
    for line in code:
        if line is None:
            rprint(None, style="bold")
        else:
            rprint("[reverse]" + line + "[/reverse]", end='')

    console.rule()


def print_rich_output(text, title=""):
    console.rule(title)
    rprint(text)
    console.rule()


def print_code(code, title=""):
    print_code_lines(code.lines, title)


def print_code_lines(lines, title=""):
    console.rule(title)
    for line in lines:
        rprint( (f"CodeLine({line.lexer_spec.lexer_cls.__name__}, "
            f"{line.has_newline}") )
        for part in line.parts:
            rprint(f"   [reverse]{part.text}[/]  -- {part.token}")

    console.rule()


def print_code_parts(parts, title=""):
    console.rule(title)
    for part in parts:
        rprint(f"   [reverse]{part.text}[/]  -- {part.token}")

    console.rule()


def compare_code_lines(lines, compare=None, title=""):
    console.rule(title)
    for index, line in enumerate(lines):
        if compare:
            try:
                console.rule(style="blue")
                for part in compare[index].parts:
                    rprint(f"   [reverse]{part.text}[/]  -- {part.token}")

            except IndexError:
                rprint("[bold red]" +
                    "!!!!! Size mismatch, only printing first arg now")
                compare = False

        for part in line.parts:
            rprint(f"   [reverse]{part.text}[/]  -- {part.token}")

    console.rule()



# renderers/formatter.py
from purdy.content import Code, Document, RenderState
from purdy.parser import token_ancestor

# =============================================================================

def conversion_handler(formatter_cls, container, exceptions):
    """Helper function for the most common case of rendering a :class:`Code`
    or :class:`Document` object.

    :param formatter_cls: Reference to a class (not an object) to instantiate
        as the formatter for each section in the container
    :param container: `Code` or `Document` object to translate
    """
    if isinstance(container, Code):
        container = Document(container)

    render_state = RenderState(container)
    for section in container:
        formatter = formatter_cls(section, exceptions)
        render_state.formatter = formatter
        section.render(render_state)

    return render_state.content

# =============================================================================

class Formatter:
    def __init__(self, section, exceptions):
        self.tag_map = {}
        self.newline = "\n"
        self.escape = lambda x:x

        self.section = section
        self.exceptions = exceptions
        self._create_tag_map()
        self.ancestor_list = section.theme.colour_map.keys()

    def _create_tag_map(self):
        for token, fg, bg, attrs in self.section.theme.values():
            self._map_tag(token, fg, bg, attrs, self.exceptions)

    def _map_tag(self, token, fg, bg, attrs, exceptions):
        raise NotImplementedError()

    def render_code_line(self, render_state, line):
        """Abstract method that gets called for each code line to be rendered

        :param render_state: contains render info and results
        :param line: the line object being rendered
        """
        raise NotImplementedError()

# ---------------------------------------------------------------------------

class StrFormatter(Formatter):
    ### Uses str.format() to create stylized output from code; assumes the
    # ._map_tag() method populated using {text} for any token text to be
    # inserted
    def render_code_line(self, render_state, line):
        for part in line.parts:
            token = token_ancestor(part.token, self.ancestor_list)
            token_text = self.escape(part.text)

            try:
                marker = self.tag_map[token]
                render_state.content += marker.format(text=token_text)
            except KeyError:
                render_state.content += token_text

        if line.has_newline:
            render_state.content += self.newline



# renderers/html.py
from html import escape as html_escape

from pygments.token import Token, Whitespace

from purdy.content import Code, Document, RenderState
from purdy.parser import HighlightOn, HighlightOff
from purdy.renderers.formatter import StrFormatter

# ===========================================================================

class HTMLFormatter(StrFormatter):
    def __init__(self, section, exceptions):
        super().__init__(section, exceptions)
        self.escape = html_escape

    def _map_tag(self, token, fg, bg, attrs, exceptions):
        if token in exceptions:
            self.tag_map[token] = exceptions[token]
            return

        # Default handling
        if not (fg or bg or attrs):
            # No formatting
            self.tag_map[token] = "{text}"
            return

        tag = ""
        if fg or bg:
            tag = '<span style="'

            if fg:
                tag += f"color: #{fg}; "

            if bg:
                tag += f"background: #{bg}; "

            tag += '">'

        if "bold" in attrs:
            tag += "<b>"

        tag += "{text}"

        if "bold" in attrs:
            tag += "</b>"

        if fg or bg:
            tag += '</span>'

        self.tag_map[token] = tag


_CODE_TAG_EXCEPTIONS = {
    Token:              '{text}',
    Whitespace:         '{text}',

    # Purdy tokens
    HighlightOn:        '<span style="background: white;">',
    HighlightOff:       '</span>',
}

# ===========================================================================

HTML_HEADER = """\
<!doctype html>
<html lang="en">
<body>
"""

HTML_FOOTER = """\
</body>
</html>
"""

def to_html(container, snippet=True):
    """Transforms tokenized content in a :class:`Code` object into a string
    representation of HTML.

    :param container: :class:`Code` or :class:`Document` object to translate
    :param snippet: When True [default] only show the code in a <div>,
        otherwise wrap it in full HTML document tags.
    """
    if isinstance(container, Code):
        container = Document(container)

    render_state = RenderState(container)

    # Header
    if not snippet:
        render_state.content += HTML_HEADER

    for section in container:
        if container.background is None:
            bg = "222222"
        else:
            bg = container.background

        render_state.content += (
            f'<div style="background :#{bg}; overflow:auto; width:auto; '
            'border:solid gray; border-width:.1em .1em .1em .8em; '
            'padding:.2em .6em;"><pre style="margin: 0; line-height:125%">'
        )

        render_state.formatter = HTMLFormatter(section, _CODE_TAG_EXCEPTIONS)
        section.render(render_state)

    if not snippet:
        render_state.content += HTML_FOOTER

    return render_state.content



# renderers/rich.py
from pygments.token import Token, Whitespace

from rich.markup import escape as rich_escape

from purdy.parser import HighlightOn, HighlightOff
from purdy.renderers.formatter import StrFormatter, conversion_handler

# ===========================================================================

class RichFormatter(StrFormatter):
    def __init__(self, section, exceptions):
        super().__init__(section, exceptions)
        self.escape = rich_escape

    def _map_tag(self, token, fg, bg, attrs, exceptions):
        if token in exceptions:
            self.tag_map[token] = exceptions[token]
            return

        # Default handling
        if not (fg or bg or attrs):
            # No formatting
            self.tag_map[token] = "{text}"
            return

        # Use f-string to inject rich markup, but raw string to insert the
        # brace brackets expected by .format_doc()
        self.tag_map[token] = f"[#{fg} {attrs}]" + r"{text}" + "[/]"


_CODE_TAG_EXCEPTIONS = {
    Token:              "{text}",
    Whitespace:         "{text}",

    # Purdy tokens
    HighlightOn:        "[reverse]{text}",
    HighlightOff:       "[/reverse]",
}

# ===========================================================================

def to_rich(container):
    """Transforms tokenized content in a :class:`Code` object into a string
    with Rich library formatting.

    :param container: `Code` or :class:`Document` object to render
    """
    return conversion_handler(RichFormatter, container, _CODE_TAG_EXCEPTIONS)



# purdy.py
import argparse

from argparse_formatter import FlexiFormatter

from purdy.content import Code
from purdy.cmds.arg_helpers import (filename_arg, general_args, motif_args,
    motif_factory)
from purdy.tui.apps import SimpleApp

# =============================================================================

DESCRIPTION = """Purdy is a library and set of command line tools for
displaying code. You can use the library to write your display animations, or
use this program as a code viewer.
"""

# =============================================================================
# Main
# =============================================================================

parser = argparse.ArgumentParser(description=DESCRIPTION,
    formatter_class=FlexiFormatter)

general_args(parser)
motif_args(parser)
filename_arg(parser)

def main():
    args = parser.parse_args()
    code = Code(args.filename, args.lexer)
    motif = motif_factory(code, args)

    app = SimpleApp(motif)
    app.run()



from copy import deepcopy
from pathlib import Path
from unittest import TestCase

from purdy.content import Code, Document, RenderState, StringSection
from purdy.parser import HighlightOn, HighlightOff, token_is_a
from purdy.renderers.plain import to_plain

# =============================================================================
# Result Constants
# =============================================================================

WRAP_SIMPLE = "\n".join([
    r'# Small file for simple parser ',
    r'testing',
    r'def simple(thing):',
    r'    """This tests',
    r'    multi-line strings"""',
    r'    return thing + "\nDone"',
    r'',
    r'simple("A string\nWith newline")',
    r'',
])

NUM_SIMPLE = "\n".join([
    r' 8 # Small file for simple parser testing',
    r' 9 def simple(thing):',
    r'10     """This tests',
    r'11     multi-line strings"""',
    r'12     return thing + "\nDone"',
    r'13 ',
    r'14 simple("A string\nWith newline")',
    r'',
])

WRAP_NUM_SIMPLE = "\n".join([
    r' 8 # Small file for simple parser ',
    r'testing',
    r' 9 def simple(thing):',
    r'10     """This tests',
    r'11     multi-line strings"""',
    r'12     return thing + "\nDone"',
    r'13 ',
    r'14 simple("A string\nWith newline")',
    r'',
])

WRAP_BIG_NUM_SIMPLE = "\n".join([
    r'10000 # Small file for simple ',
    r'parser testing',
    r'10001 def simple(thing):',
    r'10002     """This tests',
    r'10003     multi-line strings"""',
    r'10004     return thing + "\nDone"',
    r'10005 ',
    r'10006 simple("A string\nWith ',
    r'newline")',
    r'',
])

WRAP_FOLD_BIG_NUM_SIMPLE = "\n".join([
    r'10000 # Small file for simple ',
    r'parser testing',
    r'10001 def simple(thing):',
    r"10002 ⠇",
    r'10005 ',
    r'10006 simple("A string\nWith ',
    r'newline")',
    r'',
])

# =============================================================================

class BraceFormatter:
    ### Text based formatter that marks highlights with special characters
    def render_code_line(self, render_state, line):
        result = ""
        for part in line.parts:
            if token_is_a(part.token, HighlightOn):
                result += "{"
            elif token_is_a(part.token, HighlightOff):
                result += "}"
            else:
                result += part.text

        if line.has_newline:
            result += "\n"

        render_state.content += result

# =============================================================================

class TestCode(TestCase):
    def test_code_factories(self):
        text = "\n".join([str(x) for x in range(0, 10)]) + "\n"

        # Constructor as factory
        path = (Path(__file__).parent / Path("data/count.txt")).resolve()
        code = Code(path, "plain")

        self.assertEqual(10, len(code.lines))
        self.assertEqual("0", code.lines[0].parts[0].text)
        self.assertEqual("9", code.lines[9].parts[0].text)

        # Text based factory
        code = Code.text(text, "plain")

        self.assertEqual(10, len(code.lines))
        self.assertEqual("0", code.lines[0].parts[0].text)
        self.assertEqual("9", code.lines[9].parts[0].text)

        # Spawn
        spawn = code.spawn()
        self.assertEqual(code.parser, spawn.parser)
        self.assertEqual(0, len(spawn.lines))

        # Slice access
        result = code[0:2]
        self.assertEqual(2, len(result.lines))
        self.assertEqual("0", result.lines[0].parts[0].text)
        self.assertEqual("1", result.lines[1].parts[0].text)

        result = code[1]
        self.assertEqual(1, len(result.lines))
        self.assertEqual("1", result.lines[0].parts[0].text)


    def test_chunk(self):
        text = "\n".join([str(x) for x in range(0, 5)]) + "\n"
        code = Code.text(text, "plain")

        expected_lines = deepcopy(code.lines[0:3])
        result = code.chunk(3)
        self.assertEqual(expected_lines, result.lines)

        expected_lines = deepcopy(code.lines[3:5])
        result = code.chunk(3)
        self.assertEqual(expected_lines, result.lines)

        result = code.chunk(3)
        self.assertEqual([], result.lines)

        # Remaining chunk
        code.current = 0
        _ = code.chunk(3)
        expected_lines = deepcopy(code.lines[3:5])
        result = code.remaining_chunk()
        self.assertEqual(expected_lines, result.lines)

    def test_folding(self):
        # Sample text numbers 1-10 separated by newlines
        text = "\n".join([str(x) for x in range(0, 10)]) + "\n"
        code = Code.text(text, "plain")

        # Fold lines 2 through 9
        code.fold(1, 8)

        # Folding line[1], length=8 means that eight items of metadata get
        # created, validate that nothing else got created accidentally
        self.assertEqual(8, len(code.meta))
        self.assertTrue(code.meta[1].folded)
        self.assertFalse(code.meta[1].hidden)

        for x in range(2, 9):
            self.assertFalse(code.meta[x].folded)
            self.assertTrue(code.meta[x].hidden)

        # Unfold
        code.unfold(1)

        # Unfolding doesn't remove the metadata, but want to make sure nothing
        # new was added
        self.assertEqual(8, len(code.meta))

        self.assertFalse(code.meta[1].folded)
        self.assertFalse(code.meta[1].hidden)

        for x in range(2, 9):
            self.assertFalse(code.meta[x].folded)
            self.assertFalse(code.meta[x].hidden)

        # Multiple folds
        code.reset_metadata()
        code.fold(1, 1)
        code.fold(5, 2)

        # Two folds and one hidden
        self.assertEqual(3, len(code.meta))

        self.assertTrue(code.meta[1].folded)
        self.assertFalse(code.meta[1].hidden)

        self.assertTrue(code.meta[5].folded)
        self.assertFalse(code.meta[5].hidden)

        self.assertFalse(code.meta[6].folded)
        self.assertTrue(code.meta[6].hidden)

        # Unfold when multi-folded
        code.unfold(1)
        self.assertEqual(3, len(code.meta))

        self.assertFalse(code.meta[1].folded)
        self.assertFalse(code.meta[1].hidden)

        # Error handling, double-fold
        with self.assertRaises(ValueError):
            code.fold(6, 1)

        # Make sure all this mucking about hasn't created unnecessary sparse
        # entries
        self.assertEqual(3, len(code.meta))

        # Error handling, unfold something not folded
        code.reset_metadata()
        with self.assertRaises(ValueError):
            code.unfold(0)

    def test_highlight(self):
        # Tests the meta data portion of highlighting
        text = "zero\none\ntwo\nthree and a bit\nfour and a bit"
        code = Code.text(text, "plain")

        # Int index
        code.highlight(1)
        self.assertTrue(code.meta[1].highlight)
        self.assertTrue(code.is_highlighted(1))
        self.assertFalse(code.is_highlighted(2))
        code.reset_metadata()

        # Negative index
        code.highlight(-1)
        self.assertTrue(code.meta[4].highlight)
        code.reset_metadata()

        # Tuple (start, length)
        code.highlight( (1, 3) )
        self.assertTrue(code.meta[1].highlight)
        self.assertTrue(code.meta[2].highlight)
        self.assertTrue(code.meta[3].highlight)
        code.reset_metadata()

        # Tuple (negative, length)
        code.highlight( (-2, 2) )
        self.assertTrue(code.meta[3].highlight)
        self.assertTrue(code.meta[4].highlight)
        code.reset_metadata()

        # String int
        code.highlight("0")
        self.assertTrue(code.meta[0].highlight)
        code.reset_metadata()

        # String int negative
        code.highlight("-1")
        self.assertTrue(code.meta[4].highlight)
        code.reset_metadata()

        # String range
        code.highlight("0-2")
        self.assertTrue(code.meta[0].highlight)
        self.assertTrue(code.meta[1].highlight)
        self.assertTrue(code.meta[2].highlight)
        code.reset_metadata()

        # Partial Highlighting
        code.highlight("3:6,3")
        self.assertEqual( [(6, 3)], code.meta[3].highlight_partial)
        self.assertTrue(code.is_highlighted(3))
        code.reset_metadata()

        # Partial Highlighting, negative index
        code.highlight("-1:5,3")
        self.assertEqual( [(5, 3)], code.meta[4].highlight_partial)
        code.reset_metadata()

        # Mixed
        code.highlight(0, -1, "3:10,1")
        self.assertTrue(code.meta[0].highlight)
        self.assertTrue(code.meta[4].highlight)
        self.assertEqual( [(10, 1)], code.meta[3].highlight_partial)
        code.reset_metadata()

        # --- Test Highlight Off
        # Int
        code.highlight(1)
        code.highlight_off(1)
        self.assertEqual(1, len(code.meta))
        self.assertFalse(code.meta[1].highlight)
        code.reset_metadata()

        # Negative Int
        code.highlight(-1)
        code.highlight_off(-1)
        self.assertEqual(1, len(code.meta))
        self.assertFalse(code.meta[4].highlight)
        code.reset_metadata()

        # Tuple (start, length)
        code.highlight( (3, 2) )
        code.highlight_off( (3, 2) )
        self.assertEqual(2, len(code.meta))
        self.assertFalse(code.meta[3].highlight)
        self.assertFalse(code.meta[4].highlight)
        code.reset_metadata()

        # Tuple (negative start, length)
        code.highlight( (-1, 1) )
        code.highlight_off( (-1, 1) )
        self.assertEqual(1, len(code.meta))
        self.assertFalse(code.meta[4].highlight)
        code.reset_metadata()

        # Partial Highlighting
        code.highlight("3:6,3")
        code.highlight_off("3:6,3")
        self.assertEqual(1, len(code.meta))
        self.assertEqual([], code.meta[3].highlight_partial)
        code.reset_metadata()

        # String range, and goes past limit
        code.highlight("0-2")
        code.highlight_off("0-4")

        self.assertEqual(5, len(code.meta))
        for i in range(0, 5):
            self.assertFalse(code.meta[i].highlight)

        code.reset_metadata()

        # All off
        code.highlight("0-2")
        code.highlight_all_off()

        self.assertEqual(3, len(code.meta))
        for i in range(0, 3):
            self.assertFalse(code.meta[i].highlight)

        # Error handling: rejects partial that isn't on
        with self.assertRaises(ValueError):
            code.highlight_off("0:1,4")

    def test_apply_highlight(self):
        # Tests the token manipulation portion of highlighted lines
        text = "zero\none\ntwo\nthree and a bit\nfour and a bit"
        code = Code.text(text, "plain")

        code.highlight(2, "3:6,3")

        # No highlighting
        line = code._apply_highlight(0)
        self.assertEqual(line, code.lines[0])

        # Full line
        line = code._apply_highlight(2)
        self.assertTrue(token_is_a(line.parts[0].token, HighlightOn))
        self.assertTrue(token_is_a(line.parts[-1].token, HighlightOff))

        # Partial line
        line = code._apply_highlight(3)
        self.assertTrue(token_is_a(line.parts[1].token, HighlightOn))
        self.assertTrue(token_is_a(line.parts[3].token, HighlightOff))

    def test_render_highlight(self):
        text = "zero\none\ntwo\nthree and a bit\nfour and a bit"
        code = Code.text(text, "plain")
        code.highlight(1)
        code.highlight("3:6,3")

        doc = Document(code)
        rs = RenderState(doc)
        rs.formatter = BraceFormatter()

        expected = "zero\n{one}\ntwo\nthree {and} a bit\nfour and a bit"
        doc[0].render(rs)
        self.assertEqual(expected, rs.content)

    def test_wrapping(self):
        path = (Path(__file__).parent / Path("data/wrap.py")).resolve()
        code = Code(path)
        doc = Document(code)

        # Default, no wrapping case
        expected = path.read_text()
        result = to_plain(doc)
        self.assertEqual(expected, result)

        # Test wrapping at different places.
        #
        # wrap.py line 3:
        #
        # if (alpha == 3 or alpha == "a long string") and beta == 5:
        #                    |              ^=35                |
        #                    ^=20              ^=split + 20     ^=split + 20

        # Wrap once on the space at point 35
        doc.wrap = 35
        result = to_plain(doc).split("\n")
        self.assertEqual(6, len(result))
        self.assertEqual("a long ", result[2][-7:])
        self.assertEqual("string", result[3][0:6])

        # Wrap once midword at point 37, should get same result
        doc.wrap = 37
        result = to_plain(doc).split("\n")
        self.assertEqual(6, len(result))
        self.assertEqual("a long ", result[2][-7:])
        self.assertEqual("string", result[3][0:6])

        # Wrap twice at size 20
        doc.wrap = 20
        result = to_plain(doc).split("\n")
        self.assertEqual(8, len(result))
        self.assertEqual("or ", result[2][-3:])
        self.assertEqual("alpha", result[3][0:5])
        self.assertEqual("a long ", result[3][-7:])
        self.assertEqual("string", result[4][0:6])
        self.assertEqual("==", result[4][-2:])
        self.assertEqual(" 5:", result[5])

        # Test really wide wrapping
        code = Code.text("# One two three four five six seven eight nine ten eleven twelve thirteen fourteen fifteen sixteen seventeen eighteen nineteen twenty twenty-one twenty-two twenty-three twenty-four twenty-five\n")
        doc = Document(code)
        doc.wrap = 80
        result = to_plain(doc).split("\n")
        self.assertEqual(4, len(result))
        self.assertEqual("thirteen ", result[0][-9:])
        self.assertEqual("-one ", result[1][-5:])
        self.assertEqual("-five", result[2][-5:])
        self.assertEqual("", result[3])

    def test_render_variations(self):
        # Test full handling of plain.Code
        path = (Path(__file__).parent / Path("data/simple.py")).resolve()
        code = Code(path)
        doc = Document(code)
        doc.wrap = 35
        result = to_plain(doc)
        self.assertEqual(WRAP_SIMPLE, result)

        # Test wrapping with numbers
        doc.line_numbers_enabled = True
        doc.starting_line_number = 8
        result = to_plain(doc)
        self.assertEqual(WRAP_NUM_SIMPLE, result)

        # Test wrapping with numbers that bump the wrap point
        doc.starting_line_number = 10_000
        result = to_plain(doc)
        self.assertEqual(WRAP_BIG_NUM_SIMPLE, result)

        # Test wrapping and folding
        code.fold(2, 3)
        doc = Document(code)
        doc.wrap = 35
        doc.line_numbers_enabled = True
        doc.starting_line_number = 10_000
        result = to_plain(doc)
        self.assertEqual(WRAP_FOLD_BIG_NUM_SIMPLE, result)


class TestDocument(TestCase):
    def test_document_init(self):
        doc = Document()
        self.assertEqual(0, len(doc))

        text = "one"
        code1 = Code.text(text, "plain")

        doc = Document(code1)
        self.assertEqual(1, len(doc))
        self.assertEqual(code1, doc[0])

        text = "two"
        code2 = Code.text(text, "plain")

        doc = Document([code1, code2])
        self.assertEqual(2, len(doc))
        self.assertEqual(code1, doc[0])
        self.assertEqual(code2, doc[1])

    def test_mixed_content(self):
        doc = Document()
        code = Code.text("one\n", "plain")
        doc.append(code)
        s = StringSection("two\n")
        doc.append(s)
        result = to_plain(doc)
        self.assertEqual("one\ntwo\n", result)


class TestRenderState(TestCase):
    def test_line_numbers(self):
        text = "\n".join(["a" for x in range(0, 500)]) + "\n"
        code1 = Code.text(text, "plain")

        text = "\n".join(["a" for x in range(500, 1001)]) + "\n"
        code2 = Code.text(text, "plain")

        doc = Document([code1, code2])
        doc.line_numbers_enabled = True

        # Default, starting at 1
        rs = RenderState(doc)
        self.assertEqual(1, rs.line_number)
        self.assertEqual("   1 ", rs.next_line_number())
        self.assertEqual("   2 ", rs.next_line_number())

        # Starting at 10
        doc.starting_line_number = 10
        rs = RenderState(doc)
        self.assertEqual(10, rs.line_number)
        self.assertEqual("  10 ", rs.next_line_number())
        self.assertEqual("  11 ", rs.next_line_number())

        # Starting at 9_000
        doc.starting_line_number = 9999
        rs = RenderState(doc)
        self.assertEqual(9999, rs.line_number)
        self.assertEqual(" 9999 ", rs.next_line_number())
        self.assertEqual("10000 ", rs.next_line_number())

        # Gap
        self.assertEqual("      ", rs.line_number_gap())



from pathlib import Path
from unittest import TestCase

from pygments.lexers import PythonConsoleLexer
from pygments.token import Token

from purdy.content import Code
from purdy.parser import (CodeLine, CodePart, LexerSpec, Parser, PartsList,
    token_is_a, token_ancestor)

from shared import code_liner

# =============================================================================

class TestTokenComparison(TestCase):
    def test_token_is_a(self):
        # Same token case
        self.assertTrue( token_is_a(Token.Keyword, Token.Keyword) )

        # Parent token case
        self.assertTrue( token_is_a(Token.Text.Whitespace, Token.Text) )

        # Not equal case
        self.assertFalse( token_is_a(Token.Text, Token.Keyword) )

    def test_ancestor(self):
        # Perfect match case
        result = token_ancestor(Token.Text, [Token.Text])
        self.assertEqual(Token.Text, result)

        # Ancestor case
        result = token_ancestor(Token.Text.Whitespace, [Token.Text])
        self.assertEqual(Token.Text, result)

        # Default case
        result = token_ancestor(Token.Text.Whitespace, [Token.Keyword])
        self.assertEqual(Token, result)


class TestLexerSpec(TestCase):
    def test_get_spec(self):
        # Auto detect case
        expected = LexerSpec.built_ins["py"]
        result = LexerSpec.get_spec("detect", hint="data/simple.py")
        self.assertEqual(expected, result)

        # Named case
        expected = LexerSpec.built_ins["py"]
        result = LexerSpec.get_spec("py")
        self.assertEqual(expected, result)

        # LexerSpec case
        expected = LexerSpec.built_ins["py"]
        result = LexerSpec.get_spec(expected)
        self.assertEqual(expected, result)

        # Pygments Lexer Case
        result = LexerSpec.get_spec(PythonConsoleLexer)
        self.assertEqual("custom_PythonConsoleLexer", result.description)
        self.assertEqual(PythonConsoleLexer, result.lexer_cls)

        # Error handling: deal with class that isn't a Pygments Lexer, any old
        # class will do as long as it isn't a useful one
        with self.assertRaises(ValueError):
            LexerSpec.get_spec(TestCase)

    def test_find(self):
        # Test straight look-up
        lexer_spec = LexerSpec.find("repl")
        self.assertEqual(LexerSpec.built_ins["repl"], lexer_spec)

        # Test alias
        lexer_spec = LexerSpec.find("txt")
        self.assertEqual(LexerSpec.built_ins["plain"], lexer_spec)

        # Test name listing
        expected = len(LexerSpec.built_ins) + len(LexerSpec.aliases)
        count = len(LexerSpec.names)
        self.assertEqual(expected, count)

        for name in LexerSpec.built_ins.keys():
            self.assertIn(name, LexerSpec.names)

        for name in LexerSpec.aliases.keys():
            self.assertIn(name, LexerSpec.names)

        # Test bad choice
        with self.assertRaises(ValueError):
            LexerSpec.find("not_a_lexer_name")

    def test_display(self):
        # Not a full check as that really is just a doubling of the code, but
        # make sure it doesn't blow up
        self.assertIn("repl", LexerSpec.names)

        result = LexerSpec.display_choices()
        self.assertIn("repl", result)


class TestPartsList(TestCase):
    def test_partslist(self):
        parts = PartsList()
        self.assertEqual(0, parts.text_length)

        parts.append(CodePart(Token.Keyword, 'def'))
        self.assertEqual(3, parts.text_length)

        parts[0] = CodePart(Token.Keyword, 'if')
        self.assertEqual(2, parts.text_length)

        parts.insert(0, CodePart(Token.Keyword, 'def'))
        self.assertEqual(5, parts.text_length)

        del parts[1]
        self.assertEqual(3, parts.text_length)

        second = PartsList()
        second.append(CodePart(Token.Keyword, 'else'))

        parts.extend(second)
        self.assertEqual(7, parts.text_length)

        self.assertEqual("defelse", parts.all_text)


class TestCodeLine(TestCase):
    def test_codeline(self):
        plain = LexerSpec.built_ins["plain"]
        repl = LexerSpec.built_ins["repl"]

        # Check the repr doesn't blow up (read: get my coverage count up)
        line = CodeLine(plain)
        repr(line)

        # Validate error handling
        #
        # Trigger __eq__
        self.assertTrue(plain == plain)

        # Equality test: classes
        self.assertFalse(line == CodeLine(repl))

        # Equality test: newline marker
        line.has_newline = True
        self.assertFalse(line == CodeLine(plain))
        line.has_newline = False

        # Equality test: parts count
        line.parts.append(CodePart(Token.Keyword, 'def'))
        self.assertFalse(line == CodeLine(plain))

        # Equality test: equal part count but different contents
        line2 = code_liner(plain, False, CodePart(Token.Keyword, 'if'))
        self.assertFalse(line == line2)

        # Spawn
        result = line.spawn()
        self.assertEqual(line.lexer_spec, result.lexer_spec)
        self.assertEqual(line.has_newline, result.has_newline)
        self.assertEqual(0, len(result.parts))


class TestParser(TestCase):
    def test_newline_handling(self):
        parser = Parser(LexerSpec.get_spec("py"))

        # With newline
        content = "x=1\n"
        result = Code.text("", "py")
        parser.parse(content, result)
        self.assertEqual(1, len(result.lines))
        self.assertTrue(result.lines[0].has_newline)

        # Without newline
        content = "x=1"
        result = Code.text("", "py")
        parser.parse(content, result)
        self.assertEqual(1, len(result.lines))
        self.assertFalse(result.lines[0].has_newline)

    def test_simple(self):
        parser = Parser(LexerSpec.get_spec("py"))

        path = (Path(__file__).parent / Path("data/simple.py")).resolve()
        content = path.read_text()

        expected = [
            code_liner(parser.lexer_spec, True,
                CodePart(Token.Comment.Single,
                    '# Small file for simple parser testing')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(Token.Keyword, 'def'),
                CodePart(Token.Text.Whitespace, ' '),
                CodePart(Token.Name.Function, 'simple'),
                CodePart(Token.Punctuation, '('),
                CodePart(Token.Name, 'thing'),
                CodePart(Token.Punctuation, ')'),
                CodePart(Token.Punctuation, ':')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Text.Whitespace, text='    '),
                CodePart(token=Token.Literal.String.Doc,
                    text='"""This tests')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Literal.String.Doc,
                    text='    multi-line strings"""')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Text, text='    '),
                CodePart(token=Token.Keyword, text='return'),
                CodePart(token=Token.Text, text=' '),
                CodePart(token=Token.Name, text='thing'),
                CodePart(token=Token.Text, text=' '),
                CodePart(token=Token.Operator, text='+'),
                CodePart(token=Token.Text, text=' '),
                CodePart(token=Token.Literal.String.Double, text='"'),
                CodePart(token=Token.Literal.String.Escape, text='\\n'),
                CodePart(token=Token.Literal.String.Double, text='Done'),
                CodePart(token=Token.Literal.String.Double, text='"')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(Token.Text.Whitespace, '')),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Name, text='simple'),
                CodePart(token=Token.Punctuation, text='('),
                CodePart(token=Token.Literal.String.Double, text='"'),
                CodePart(token=Token.Literal.String.Double, text='A string'),
                CodePart(token=Token.Literal.String.Escape, text='\\n'),
                CodePart(token=Token.Literal.String.Double,
                    text='With newline'),
                CodePart(token=Token.Literal.String.Double, text='"'),
                CodePart(token=Token.Punctuation, text=')')
            ),
        ]

        result = Code.text("", "py")
        parser.parse(content, result)
        self.assertEqual(expected, result.lines)

    def test_blank_lines(self):
        parser = Parser(LexerSpec.get_spec("py"))

        content = "   \n\nx=1\n"
        expected = [
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Text, text='   ')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Text.Whitespace, text='')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Name, text='x'),
                CodePart(token=Token.Operator, text='='),
                CodePart(token=Token.Literal.Number.Integer, text='1')
            ),
        ]

        result = Code.text("", "py")
        parser.parse(content, result)
        self.assertEqual(expected, result.lines)

    def test_multiline_output(self):
        parser = Parser(LexerSpec.get_spec("con"))

        path = (Path(__file__).parent / Path("data/curl.con")).resolve()
        content = path.read_text()

        expected = [
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Generic.Output, text='')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Generic.Prompt, text='$ '),
                CodePart(token=Token.Text, text='curl'),
                CodePart(token=Token.Text.Whitespace, text=' '),
                CodePart(token=Token.Text, text='--include'),
                CodePart(token=Token.Text.Whitespace, text=' '),
                CodePart(token=Token.Text,
                    text='http://127.0.0.1:8000/redirect/')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Generic.Output,
                    text='HTTP/1.1 302 Found')
            ),
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Generic.Output,
                    text='Date: Tue, 21 Apr 2020 19:31:07 GMT')
            ),
        ]

        result = Code.text("", "con")
        parser.parse(content, result)
        self.assertEqual(expected, result.lines)

    def test_after_newline(self):
        # Some lexers aren't line oriented so you can end up with stuff after
        # a \n character
        content = "<html>\n     <!-- comment -->"
        parser = Parser(LexerSpec.get_spec("html"))

        expected = [
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Punctuation, text='<'),
                CodePart(token=Token.Name.Tag, text='html'),
                CodePart(token=Token.Punctuation, text='>')
            ),
            code_liner(parser.lexer_spec, False,
                CodePart(token=Token.Text, text='     '),
                CodePart(token=Token.Comment.Multiline,
                    text='<!-- comment -->')
            ),
        ]

        result = Code.text("", "html")
        parser.parse(content, result)
        self.assertEqual(expected, result.lines)

    def test_empty_token(self):
        # Some lexers spit out empty tokens, purdy should ignore them
        parser = Parser(LexerSpec.get_spec("repl"))
        content = """Traceback (most recent call last):\n>>> """

        expected = [
            code_liner(parser.lexer_spec, True,
                CodePart(token=Token.Generic.Traceback,
                    text='Traceback (most recent call last):')
            ),
            code_liner(parser.lexer_spec, False,
                CodePart(token=Token.Generic.Prompt, text='>>> ')
            ),
        ]

        result = Code.text("", "repl")
        parser.parse(content, result)
        self.assertEqual(expected, result.lines)



# renderers/textual.py
from pygments.token import Token, Whitespace
from textual.content import Content

from purdy.parser import HighlightOn, HighlightOff, token_ancestor
from purdy.renderers.formatter import conversion_handler, Formatter

# ===========================================================================

class TextualFormatter(Formatter):
    def _map_tag(self, token, fg, bg, attrs, exceptions):
        if token in exceptions:
            self.tag_map[token] = exceptions[token]
            return

        # Default handling
        if not (fg or bg or attrs):
            # No formatting
            self.tag_map[token] = "$text"
            return

        # Use f-string to inject rich markup, but raw string to insert the
        # brace brackets expected by .format_doc()
        self.tag_map[token] = f"[#{fg} {attrs}]$text[/]"

    def render_code_line(self, render_state, line):
        # Textual really doesn't like piecemeal creation of content or
        # strings, and they way the code works elsewhere you can just append a
        # close attr, but here you can't
        #
        # You need to construct a single string for the Content.from_markup
        # method, which means having to be a bit hacky, creating then
        # re-parsing the individual pieces

        part_map = {}
        counter = 0
        markup = ""

        for part in line.parts:
            token = token_ancestor(part.token, self.ancestor_list)

            name = f"text_{counter}"
            dname = "$" + name

            try:
                # Get the tag from the general map and replace $text with the
                # counted version
                marker = self.tag_map[token]
                marker = marker.replace("$text", dname)
            except KeyError:
                # No map tags, use just the placeholder for the content
                marker = dname

            # Store the text for Content's kwargs and update our markup
            # string
            part_map[name] = part.text
            markup += marker

            counter += 1

        if line.has_newline:
            markup += self.newline

        # Now get Textual to render that mess
        render_state.content += Content.from_markup(markup, **part_map)

    def part_to_content(self, token, value):
        token = token_ancestor(token, self.ancestor_list)
        part_map = {
            "text": value,
        }

        try:
            # Use the tag from the general map
            markup = self.tag_map[token]
        except KeyError:
            # No map tags, use just the placeholder for the content
            markup = "$text"

        # Get Textual to render that mess
        return Content.from_markup(markup, **part_map)


_CODE_TAG_EXCEPTIONS = {
    Token:              "$text",
    Whitespace:         "$text",

    # Purdy tokens
    HighlightOn:        "[reverse]$text",
    HighlightOff:       "[/reverse]",
}

# ===========================================================================

def to_textual(container):
    """Transforms tokenized content in a :class:`Code` or :class:`Document`
    object into a string with Textual library formatting.

    :param container: :class:`Code` or :class:`Document` object to translate
    """
    return conversion_handler(TextualFormatter, container, _CODE_TAG_EXCEPTIONS)



# subpurdy.py
import argparse

from argparse_formatter import FlexiFormatter
from rich.console import Console

from purdy.cmds.arg_helpers import (filename_arg, general_args, no_colour_arg,
    doc_args, document_factory)
from purdy.renderers.html import to_html
from purdy.renderers.rich import to_rich
from purdy.renderers.rtf import to_rtf
from purdy.scribe import print_code_lines

# =============================================================================

DESCRIPTION = """Purdy is a library and set of command line tools for
displaying code. You can use the library to write your display animations, or
use the subcommands of this program for preset usages. The 'purdy' command
uses the Textual library to display a colourized version of your code in the
console, while 'subpurdy' command is a set of utilities for printing different
styles of code to the screen.
"""

console = Console(highlight=False)
rprint = console.print

# =============================================================================
# Subcommands
# =============================================================================

def tokens(args):
    ### 'tokens' sub-command: prints lines and tokens to screen
    doc = document_factory(args)
    print_code_lines(doc[0].lines, title_enabled=False, no_colour=args.nocolour)


def ansi(args):
    ### 'ansi' sub-command: prints content with ANSI colour highlighting
    doc = document_factory(args)
    output = to_rich(doc)
    rprint(output)


def html(args):
    ### 'html' sub-command: prints content as an HTML div
    doc = document_factory(args)
    output = to_html(doc, not args.fullhtml)
    print(output)


def rtf(args):
    ### 'rtf' sub-command: prints content in RTF format
    doc = document_factory(args, "rtf")
    output = to_rtf(doc)
    print(output)

# =============================================================================
# Main
# =============================================================================

parser = argparse.ArgumentParser(description=DESCRIPTION,
    formatter_class=FlexiFormatter)

# === Common flags ===
general_args(parser)

# === Subcommands ===
subparsers = parser.add_subparsers(title="subcommands", dest="command")
subparsers.required = True

# --- tokens cmd
sub = subparsers.add_parser("tokens", help=("Prints out each line in a file "
    "with the corresponding tokens indented beneath it"))
no_colour_arg(sub)
sub.set_defaults(func=tokens)

# --- ansi cmd
sub = subparsers.add_parser("ansi", help=("Prints code with colourized ANSI "
    "results in your terminal"))
doc_args(sub)
sub.set_defaults(func=ansi)

# --- html cmd
sub = subparsers.add_parser("html", help="Prints code as HTML")
sub.add_argument("--fullhtml", help=("By default only a div with the code is "
    "shown. This flag causes a full HTML doc."), action="store_true")
doc_args(sub)
sub.set_defaults(func=html)

# --- rtf cmd
sub = subparsers.add_parser("rtf", help="Prints code as RTF")
doc_args(sub)
sub.set_defaults(func=rtf)

# === Positional filename argument common to all subs
filename_arg(parser)

def main():
    args = parser.parse_args()
    args.func(args)



# purdy.tui.animate.py
import asyncio
import typing

from dataclasses import dataclass
from enum import Enum

from textual.content import Content as TContent

# ===========================================================================
# Cells and Cell Management
#
# These define the possible steps that can happen during screen animation

def _content_to_repr(content):
    if isinstance(content, TContent):
        return f"Content({repr(content.markup)}), "

    return f"str({content}), "


class UndoableCell:
    def __init__(self):
        self.forwards_map = {}
        self.backwards_map = {}

    def forwards(self):
        for codebox, after in self.forwards_map.items():
            codebox.update(after)

    def backwards(self):
        for codebox, back in self.backwards_map.items():
            codebox.update(back)


class Cell(UndoableCell):
    def __init__(self, codebox, after):
        super().__init__()
        self.codebox = codebox
        self.forwards_map[self.codebox] = after

    @property
    def after(self):
        return self.forwards_map[self.codebox]

    @property
    def before(self):
        try:
            return self.backwards_map[self.codebox]
        except:
            return ""

    def __repr__(self):
        result = f"Cell(codebox={self.codebox.id}, after="
        result += _content_to_repr(self.after)
        result += " before="
        result += _content_to_repr(self.before)

        return result


@dataclass
class MoveByCell:
    codebox: typing.Any
    amount: int


@dataclass
class PauseCell:
    pause: float


class WaitCell:
    pass


@dataclass
class DebugCell:
    content: str


class TransitionCell(UndoableCell):
    def __init__(self, codebox, after, effect_cls, effect_kwargs):
        super().__init__()
        self.codebox = codebox
        self.forwards_map[self.codebox] = after
        self.effect_cls = effect_cls
        self.effect_kwargs = effect_kwargs

    def __repr__(self):
        result = f"TransitionCell(codebox={self.codebox.id}, after="
        result += _content_to_repr(self.forwards_map[self.codebox])
        result += " before="
        result += _content_to_repr(self.before)

        return result

    @property
    def before(self):
        try:
            return self.backwards_map[self.codebox]
        except:
            return ""

    async def run(self, animation_controller):
        container = self.codebox.widget.container
        overlay = self.codebox.widget.overlay

        self.effect = self.effect_cls(container, callback=self.callback,
            post_callback=animation_controller.transition_complete,
            **self.effect_kwargs)
        overlay.mount(self.effect)
        self.worker = animation_controller.app.run_worker(self.effect.run(),
            exclusive=True)

    async def cancel(self):
        self.worker.cancel()
        self.effect.remove()
        self.codebox.update(self.after)

    async def callback(self):
        self.forwards()


class ScreenTransitionCell(UndoableCell):
    def __init__(self, control, changes, effect_cls, effect_kwargs):
        super().__init__()
        self.control = control
        self.effect_cls = effect_cls
        self.effect_kwargs = effect_kwargs

        for codebox, after in changes.items():
            self.forwards_map[codebox] = after

    async def run(self, animation_controller):
        container = self.control.container.inner
        overlay = self.control.container.overlay

        self.effect = self.effect_cls(container, callback=self.callback,
            post_callback=animation_controller.transition_complete,
            **self.effect_kwargs)
        overlay.mount(self.effect)
        self.worker = animation_controller.app.run_worker(self.effect.run(),
            exclusive=True)

    async def cancel(self):
        self.worker.cancel()
        self.effect.remove()
        self.forwards()

    async def callback(self):
        self.forwards()

# ===========================================================================
# Animation Controller
# ===========================================================================

class AnimationController:
    class State(Enum):
        PAUSE = "p"
        TRANSITION = "t"
        WAIT = "w"
        DONE = "d"

    def __init__(self, app):
        self.app = app
        self.current = None
        self.worker = None
        self.wait_state = None

        # Loop through and calculate the "before" states, also mark the first
        # Wait cell in our collection
        self.first_wait = None
        before_cells = {}
        for index, cell in enumerate(cell_list):
            if isinstance(cell, WaitCell) and self.first_wait is None:
                # Found first wait
                self.first_wait = index

            if isinstance(cell, UndoableCell):
                for codebox in cell.forwards_map.keys():
                    if codebox in before_cells:
                        prev = before_cells[codebox]
                        cell.backwards_map[codebox] = prev.forwards_map[codebox]

                    before_cells[codebox] = cell

    # --- Coroutines
    async def pause_running(self, amount):
        await asyncio.sleep(amount)
        self.wait_state = None
        await self.forwards()

    async def transition_complete(self):
        # Called by transition effect when it is done
        self.wait_state = None
        await self.forwards()

    async def forwards(self):
        if self.current is not None and self.current >= len(cell_list):
            # Ignore call, we're done
            return

        if self.wait_state in [self.State.PAUSE, self.State.TRANSITION]:
            # Ignore forwards during pause or animations, make them use skip
            return

        while True:
            if self.current is None:
                # Haven't started yet, initialize
                self.current = -1

            # Current cell is the last one to do something, increment
            self.current += 1
            if self.current >= len(cell_list):
                # All cells processed, can exit the worker
                self.state = self.State.DONE
                return

            cell = cell_list[self.current]

            if isinstance(cell, DebugCell):
                # Ignore it
                continue
            elif isinstance(cell, MoveByCell):
                print("I think I'm moving by", cell.amount)
                cell.codebox.widget.vs.scroll_relative(y=cell.amount)
                continue
            elif isinstance(cell, PauseCell):
                # Pause directive, kick off the timer and leave
                self.wait_state = self.State.PAUSE
                self.worker = self.app.run_worker(
                    self.pause_running(cell.pause))
                return
            elif isinstance(cell, TransitionCell):
                self.wait_state = self.State.TRANSITION
                await cell.run(self)
                return
            elif isinstance(cell, ScreenTransitionCell):
                self.wait_state = self.State.TRANSITION
                await cell.run(self)
                return
            elif isinstance(cell, WaitCell):
                # Wait until the next time forwards is called
                self.wait_state = self.State.WAIT
                return

            # Perform cell action
            self.wait_state = None
            cell.forwards()

    async def skip(self):
        if self.current is not None and self.current >= len(cell_list):
            return

        if self.wait_state == self.State.PAUSE:
            self.worker.cancel()
            self.wait_state = None

        if self.wait_state == self.State.TRANSITION:
            cell = cell_list[self.current]
            await cell.cancel()
            self.wait_state = None

        while True:
            if self.current is None:
                # Haven't started yet, initialize
                self.current = -1

            # Current cell is the last one to do something, increment
            self.current += 1
            if self.current >= len(cell_list):
                # All cells processed, can exit the worker
                self.state = self.State.DONE
                return

            cell = cell_list[self.current]
            if isinstance(cell, DebugCell):
                # Ignore it
                continue
            elif isinstance(cell, MoveByCell):
                cell.codebox.widget.vs.scroll_relative(y=cell.amount,
                    animate=False)
                continue
            elif isinstance(cell, PauseCell):
                # Ignore pauses during skip
                continue
            if isinstance(cell, (ScreenTransitionCell, TransitionCell)):
                # Skip animation, but perform replacement
                cell.forwards()
                continue
            elif isinstance(cell, WaitCell):
                # Skipping done
                self.wait_state = self.State.WAIT
                return

            # Perform cell actions
            self.wait_state = None
            cell.forwards()

    async def backwards(self):
        if self.current is None:
            # Can't go backwards when nothing ever started
            return

        if self.wait_state == self.State.PAUSE:
            self.worker.cancel()
        elif self.wait_state == self.State.DONE:
            # We were done, undo to the previous wait
            self.current = len(cell_list) - 1

        start = self.current - 1
        if start < 0:
            # First state was a Wait, that's weird, let's ignore it
            return

        end = 0
        if self.first_wait is not None:
            end = self.first_wait

        for self.current in range(start, end - 1, -1):
            cell = cell_list[self.current]
            if isinstance(cell, DebugCell):
                # Ignore it
                continue
            elif isinstance(cell, MoveByCell):
                scroll_by = -1 * cell.amount
                cell.codebox.widget.vs.scroll_relative(y=scroll_by)
                continue
            if isinstance(cell, PauseCell):
                # Ignore pauses during backwards
                continue
            elif isinstance(cell, WaitCell):
                # Done moving backwards, set to next cell
                self.wait_state = self.State.WAIT
                return

            # Perform cell actions. TransitionCells get handled here as
            # well, as they don't animate going backwards
            self.wait_state = None
            cell.backwards()

# ===========================================================================

# Module level container for cell actions
cell_list = []



# purdy.tui.purdybox.py
from textual_transitions import Curtain

from purdy.content import Document
from purdy.renderers.textual import to_textual
from purdy.tui import animate
from purdy.tui.codebox import TextSection
from purdy.tui.widgets import PurdyContainer

# =============================================================================

class PurdyBox:
    def __init__(self, row_specs, max_height=None):
        self.row_specs = row_specs
        self.rows = []

        self.container = PurdyContainer(self, row_specs, max_height)

    def transition(self, changes=None, speed=1):
        box_changes = {}
        if changes is None:
            changes = {}

        for codebox, content in changes.items():
            if content is None:
                codebox.doc = Document()
                box_changes[codebox] = ""
            else:
                if isinstance(content, str):
                    codebox.doc = Document(TextSection(content))
                else:
                    codebox.doc = Document(content)

                after = to_textual(codebox.doc)
                box_changes[codebox] = after

        tx = animate.ScreenTransitionCell(self, box_changes, Curtain,
            {"seconds":speed})
        animate.cell_list.append(tx)
        return self



"""
Parser
=======

Contains methods and classes to manage parsing of code.
"""
from dataclasses import dataclass, field
from pathlib import Path

from pygments.lexer import Lexer as Pygments_Lexer
from pygments.lexers import (BashSessionLexer, CssLexer, JsonLexer,
    NodeConsoleLexer, PythonConsoleLexer, PythonLexer, TOMLLexer)
from pygments.lexers.data import YamlLexer
from pygments.lexers.markup import MarkdownLexer, RstLexer
from pygments.lexers.templates import HtmlDjangoLexer
from pygments.token import Generic, String, Token

from purdy.lexers import DollarBashSessionLexer, NewlineLexer

# =============================================================================
# Token Utilities
# =============================================================================

# Create new token types using Pygment's tuple magic
HighlightOn = Token.HighlightOn
HighlightOff = Token.HighlightOff
Fold = Token.Fold
LineNumber = Token.LineNumber

# -----------------------------------------------------------------------------

def token_is_a(token1, token2):
    """Returns true if token1 is the same type as or a child type of token2"""
    if token1 == token2:
        return True

    parent = token1.parent
    while(parent != None):
        if parent == token2:
            return True

        parent = parent.parent

    return False


def token_ancestor(token, ancestor_list):
    """Tokens are hierarchical, in some situations you need to translate a
    token into one from a known list, e.g. turning a
    "Token.Literal.Number.Integer" into a "Number". This method takes a token
    and a list of approved ancestors and attempts to make the map. If no
    ancestor is found then a generic "Token" object is returned

    :param token: token to translate into an approved ancestor
    :param ancestor_list: list of approved ancestor tokens
    """
    if token in ancestor_list:
        return token

    # token not in the approved list, search its ancestors
    token = token.parent
    while(token != None):
        if token in ancestor_list:
            return token

        token = token.parent

    # something went wrong with our lookup, return the default
    return Token


# =============================================================================
# Lexer Manager
# =============================================================================

@dataclass
class LexerSpec:
    """LexerSpec wraps a Pygments Lexer and contains data that changes the
    behaviour of the display of the code.

    :param description: description text about what this LexerSpec is for
    :param lexer_cls: The Pygments Lexer class used for parsing
    :param console: True if the content is a REPL or console, False otherwise.
        This setting effects how animations and certain kinds of output are
        displayed
    :param category: Classification to use for the colourization and rendering
        map
    """
    description: str
    lexer_cls: object
    console: bool
    category: str

    @classmethod
    def get_spec(cls, identifier, hint=''):
        """Return a `LexerSpec` object based on a given identifier.

        :param identifier: An indicator as to what underlying lexer to use. It
            can be the string "detect" to attempt to auto-detect the
            appropriate lexer, a string corresponding to one of the built-in
            :class:`LexerSpec` objects, a :class:`LexerSpec` itself, or a
            Pygments :class:`pygments.lexers.Lexer` class. When a Pygments
            Lexer is provided it is assumed to be for code and not in console
            mode.
        :param hint: when using identifier="detect", this provides information
            for doing the detection
        """
        match identifier:
            case "detect":
                # Assume the hint is a file name
                path = Path(hint)
                return LexerSpec.find(path.suffix[1:])
            case str(name):
                return LexerSpec.find(name)
            case LexerSpec():
                return identifier
            case _:
                if not issubclass(identifier, Pygments_Lexer):
                    raise ValueError("Could not determine Parser type")

                name = 'custom_' + identifier.__name__
                return LexerSpec(name, identifier, False, 'code')

    @classmethod
    def find(cls, name):
        name = name.lower()
        if name in cls.aliases:
            return cls.built_ins[cls.aliases[name]]

        try:
            return cls.built_ins[name]
        except KeyError:
            raise ValueError("Invalid LexerSpec name. Choices are:" +
                ",".join(cls.names))

    @classmethod
    def display_choices(cls):
        result = []
        for key, value in cls.built_ins.items():
            result.append(f"'{key}': {value.description}")

        return ", ".join(result)


LexerSpec.built_ins = {
    'py': LexerSpec('Python 3 Source', PythonLexer, False, 'code'),
    'repl': LexerSpec('Interactive Python 3 console', PythonConsoleLexer,
        True, 'code'),
    'con': LexerSpec('Interactive bash console', BashSessionLexer, True,
        'code'),

    'css': LexerSpec('CSS', CssLexer, False, 'doc'),
    'dbash': LexerSpec('Interactive bash Console with a dollar-sign prompt',
        DollarBashSessionLexer, True, 'code'),
    'html': LexerSpec('HTML/Django/Jinja', HtmlDjangoLexer, False, 'xml'),
    'json': LexerSpec('JSON', JsonLexer, False, 'doc'),
    'md': LexerSpec('Markdown Doc', MarkdownLexer, False, 'doc'),
    'node': LexerSpec('Interactive JavaScript Node.js Console',
        NodeConsoleLexer, True, 'code'),
    'plain': LexerSpec('Plain text, no parsing', NewlineLexer, False, 'doc'),
    'rst': LexerSpec('RST Doc', RstLexer, False, 'doc'),
    'toml': LexerSpec('TOML', TOMLLexer, False, 'doc'),
    'yaml': LexerSpec('YAML Doc', YamlLexer, False, 'doc'),
}

LexerSpec.aliases = {
    "htm": "html",
    "tcss": "css",
    "txt": "plain",
    "yml": "yaml",
}

LexerSpec.names = list(LexerSpec.built_ins.keys()) + \
    list(LexerSpec.aliases.keys())

# =============================================================================
# Code Token Classes
# =============================================================================

@dataclass
class CodePart:
    token: Token
    text: str

    def __repr__(self):
        return f"CodePart(id={id(self)}, token={self.token}, text={self.text})"


class PartsList(list):
    """Subclass of list that tracks the length of the text fields within the
    :class:`CodePart` objects inside of it."""
    def __init__(self, iterable=None):
        super().__init__()
        self.text_length = 0
        if iterable:
            for item in iterable:
                self.append(item)

    def __setitem__(self, index, value):
        # Replacing a value, remove the old length, add the new one
        self.text_length -= len(self[index].text)
        super().__setitem__(index, value)
        self.text_length += len(value.text)

    def __delitem__(self, index):
        self.text_length -= len(self[index].text)
        super().__delitem__(index)

    def append(self, value):
        super().append(value)
        self.text_length += len(value.text)

    def insert(self, index, value):
        super().insert(index, value)
        self.text_length += len(value.text)

    def extend(self, other):
        super().extend(other)
        for part in other:
            self.text_length += len(part.text)

    @property
    def all_text(self):
        # Mostly for debugging, prints out all the text as one blob, ignoring
        # tokens
        return "".join([part.text for part in self])


# Cursor used in typewriting animations
CURSOR_CHAR = "\u2588"
CURSOR = CodePart(token=Generic.Text, text=CURSOR_CHAR)


@dataclass
class CodeLine:
    lexer_spec: LexerSpec
    parts: PartsList = field(default_factory=PartsList)
    has_newline: bool = False

    def spawn(self):
        """Returns a new :class:`CodeLine` with the same settings as this one
        but an empty parts list."""
        return CodeLine(self.lexer_spec, has_newline=self.has_newline)

    def __repr__(self):
        return (f"CodeLine(id={id(self)}, lexer_spec={self.lexer_spec.lexer_cls.__name__}, "
            f"parts={self.parts}, has_newline={self.has_newline})")

    def __eq__(self, compare):
        if self.lexer_spec.lexer_cls != compare.lexer_spec.lexer_cls:
            return False

        if self.has_newline != compare.has_newline:
            return False

        if len(self.parts) != len(compare.parts):
            return False

        for num, part in enumerate(self.parts):
            if part != compare.parts[num]:
                return False

        return True

    def is_prompt(self):
        try:
            prompt_prefix = token_is_a(self.parts[0].token, Generic.Prompt)

            # Only prompt
            if len(self.parts) == 1 and prompt_prefix:
                return True

            # Prompt and cursor
            if len(self.parts) == 2 and prompt_prefix and \
                    self.parts[1] == CURSOR:
                return True
        except IndexError:
            pass

        return False

    def compress(self):
        """Looks for sequence of parts in a row with the same token type and
        merges them.
        """
        output = PartsList()
        previous = None
        for part in self.parts:
            if previous and output[-1].token == part.token:
                output[-1].text += part.text
            else:
                output.append(part)
                previous = part

        self.parts = output

# =============================================================================
# Parsing
# =============================================================================

class Parser:
    """Parser is responsible for parsing code and returning a :class:`Code`
    object containing :class:`CodeLine` objects. You configure a Parser with a
    :class:`LexerSpec` which you can get through the
    :method:`LexerSpec.get_spec` method.

    :param lexer_spec: a :class:`LexerSpec` object
    """
    def __init__(self, lexer_spec):
        self.lexer_spec = lexer_spec

    def parse(self, content, code_obj):
        """Parses the given content using the class's associated lexer.

        :param content: String to be parsed
        :param code_obj: A :class:`Code` object to add the resulting
            `CodeLine` objects into
        """
        # Instantiate the lexer so that it doesn't remove starting newlines,
        # just keeps it the way the content is.
        #
        # NB: it would be nice to also use "ensurenl=False" to stop the stupid
        # appending newlines, but that breaks a whole bunch of the lexers
        lexer = self.lexer_spec.lexer_cls(stripnl=False)
        self.line = CodeLine(self.lexer_spec)
        for token_type, text in lexer.get_tokens(content):
            if text.startswith('\n'):
                self._newline_handler(token_type, code_obj)
                if len(text) > 1:
                    # something came after the \n, handle it
                    #
                    # Example: HTML lexer isn't line oriented, so a \n
                    # followed by in indent is seeing as one chunk of text
                    self._string_handler(token_type, text[1:], code_obj)
            elif text == '':
                # tokenizer sometimes puts in empty stuff, skip it
                #
                # Example: Python Console Lexer and Traceback output
                continue
            elif token_is_a(token_type, String) and '\n' in text:
                self._string_handler(token_type, text, code_obj)
            else:
                self._default_handler(token_type, text, code_obj)

        # Pygments adds newlines because some of the lexers need it there,
        # get rid of it
        last_line = code_obj.lines[-1]
        if len(last_line.parts) == 1 and last_line.parts[0].text == '':
            # Got a garbage extra line, remove it
            del code_obj.lines[-1]

        # Fix the last lines newline state
        try:
            code_obj.lines[-1].has_newline = content[-1] == "\n"
        except IndexError:
            # Empty container case can be ignored
            pass

    def _newline_handler(self, token_type, code_obj):
        # hit a CR, time to create a new line
        if not self.line.parts:
            self.line = CodeLine(self.lexer_spec,
                PartsList([CodePart(token_type, '')]))

        self.line.has_newline = True
        code_obj.lines.append(self.line)

        # reset to start the next set of tokens
        self.line = CodeLine(self.lexer_spec)

    def _string_handler(self, token_type, text, code_obj):
        # String tokens may be multi-line
        for row in text.splitlines(True):
            part = CodePart(token_type, row.rstrip('\n'))
            self.line.parts.append(part)

            if row[-1] == '\n':
                self.line.has_newline = True
                code_obj.lines.append(self.line)
                self.line = CodeLine(self.lexer_spec)

    def _default_handler(self, token_type, text, code_obj):
        if text[-1] == '\n':
            # there is a \n at the end of the text, need to strip it out then
            # start a newline using the same token
            #
            # Example: this happens with multi-line output of a bash console
            # session
            part = CodePart(token_type, text.rstrip('\n'))
            self.line.parts.append(part)

            # text caused a CR, create a new line object
            self.line.has_newline = True
            code_obj.lines.append(self.line)

            # reset to start the next group of tokens
            self.line = CodeLine(self.lexer_spec)
        else:
            part = CodePart(token_type, text)
            self.line.parts.append(part)



# purdy.tui.codebox.py
import random
from copy import deepcopy
from dataclasses import dataclass

from textual_transitions import Curtain

from purdy.content import Code, Document, RenderState
from purdy.renderers.textual import (TextualFormatter, to_textual,
    _CODE_TAG_EXCEPTIONS)
from purdy.tui import animate
from purdy.tui.widgets import CodeWidget
from purdy.tui.tui_content import EscapeText, TextSection
from purdy.tui.typewriter import code_typewriterize, textual_typewriterize

# =============================================================================
# Specs Used to Define The Layout
# =============================================================================

@dataclass
class BoxSpec:
    """Used to describe each text area in the app that will be showing
    code.

    :param width: relative width based on other boxes in the row. For example,
        row_specs=[Box_Spec(2), BoxSpec(1)] results in a single row where the
        first box takes up 2/3rds of the space.

    :param line_number: Starting line number for code displayed in the box.
        Defaults to None

    :param auto_scroll: True to scroll down when content gets added

    :param border: A string specifying which borders are on for this box.
        Expects the letters "t", "b", "l", "r" in any order for turning on the
        top, bottom, left, and right borders respectively. Defaults to no
        borders.
    """
    width: int
    line_number: int = None
    auto_scroll: bool = False
    border: str = ""
    title: str = ""


@dataclass
class RowSpec:
    """Container for a row of :class:`BoxSpec` objects to describe a row in
    the display grid.

    :param height: relative height of this row in comparison to others in the
        grid. For example [RowSpec(2, ...), RowSpec(1,...)] produces two rows
        with the first taking up 2/3rds of the height of the screen.

    :param boxes: a list of :class:`BoxSpec` objects in this row
    """
    height: int
    boxes: list

# =============================================================================
# Code Abstractions
# =============================================================================

class CodeBox:
    def __init__(self, id, row_spec, box_spec):
        self.id = id
        self.box_spec = box_spec
        self.last_after = None

        self.widget = CodeWidget(border=box_spec.border, title=box_spec.title)
        self.widget.styles.row_span = row_spec.height
        self.widget.styles.column_span = box_spec.width

        self.doc = Document()

    def __repr__(self):
        return f"CodeBox({self.id})"

    def update(self, content):
        self.widget.code_display.update(content)

        if self.box_spec.auto_scroll:
            # Scroll down without any animation, we're already near the bottom
            self.widget.vs.scroll_end(animate=False)

    def _process_content(self, content):
        if isinstance(content, Code):
            self.last_parser = content.parser

    # === Animation Actions
    #
    # --- Editing Actions
    def append(self, content):
        if isinstance(content, str):
            # If there is already a text section add to it, otherwise create
            # one
            if len(self.doc) > 0 and isinstance(self.doc[-1], TextSection):
                self.doc[-1].lines.append(content)
            else:
                self.doc.append(TextSection(content))
        else:
            # Must be Code
            self.doc.append(content)

        after = to_textual(self.doc)
        animate.cell_list.append(animate.Cell(self, after))
        return self

    def clear(self):
        self.doc = Document()
        animate.cell_list.append(animate.Cell(self, ""))
        return self

    def replace(self, content):
        if isinstance(content, str):
            self.doc = Document(TextSection(content))
        else:
            self.doc = Document(content)

        after = to_textual(self.doc)
        animate.cell_list.append(animate.Cell(self, after))
        return self

    def transition(self, content=None, speed=1):
        if content is None:
            self.doc = Document()
            after = ""
        else:
            if isinstance(content, str):
                self.doc = Document(TextSection(content))
            else:
                self.doc = Document(content)

            after = to_textual(self.doc)

        tx = animate.TransitionCell(self, after, Curtain, {"seconds":speed})
        animate.cell_list.append(tx)
        return self

    # --- Typewriter actions
    def _pre_render(self, future_length=0):
        render_state = RenderState(self.doc, future_length=future_length)

        for section in self.doc:
            formatter = TextualFormatter(section, _CODE_TAG_EXCEPTIONS)
            render_state.formatter = formatter
            section.render(render_state)

        return render_state

    def typewriter(self, code, skip_comments=True, skip_whitespace=True,
            delay=0.13, delay_variance=0.03):
        if not isinstance(code, Code):
            raise ValueError("Code only! Use text_typewriter instead")

        render_state = self._pre_render(len(code.lines))
        steps = code_typewriterize(render_state, code, skip_comments,
            skip_whitespace)

        for step, state in steps:
            if isinstance(step, animate.WaitCell):
                animate.cell_list.append(step)
            else:
                after = deepcopy(render_state.content) + step
                animate.cell_list.append(animate.Cell(self, after))

                match state:
                    case "P":
                        self.pause(delay, delay_variance)
                    case "W":
                        animate.cell_list.append(animate.WaitCell())
                    # fall-through has no action

        # Final value should be the whole thing
        self.doc.append(code)
        render_state = self._pre_render()
        animate.cell_list.append(animate.Cell(self, render_state.content))
        return self

    def text_typewriter(self, content, delay=0.13, delay_variance=0.03):
        if isinstance(content, (str, EscapeText)):
            section = TextSection(content)
        else:
            section = content

        render_state = self._pre_render(len(section.lines))
        steps = textual_typewriterize(render_state, section)

        for step in steps:
            after = deepcopy(render_state.content) + step
            animate.cell_list.append(animate.Cell(self, after))

            self.pause(delay, delay_variance)

        # Final value should be the whole thing
        self.doc.append(section)
        render_state = self._pre_render()
        animate.cell_list.append(animate.Cell(self, render_state.content))
        return self

    # --- GUI Actions
    def move_by(self, amount):
        animate.cell_list.append(animate.MoveByCell(self, amount))
        return self

    def debug(self, content):
        animate.cell_list.append(animate.DebugCell(content))
        return self

    # --- Timing actions
    def pause(self, pause, pause_variance=None):
        if pause_variance is not None:
            pause = random.uniform(pause, pause + pause_variance)

        animate.cell_list.append(animate.PauseCell(pause))
        return self

    def wait(self):
        animate.cell_list.append(animate.WaitCell())
        return self

    # --- Formatting actions
    def set_numbers(self, starting_num):
        if starting_num is None:
            self.doc.line_numbers_enabled = False
            self.doc.starting_line_number = 1
        else:
            self.doc.line_numbers_enabled = True
            self.doc.starting_line_number = starting_num

        after = to_textual(self.doc)

        animate.cell_list.append(animate.Cell(self, after))
        return self

    # --- Highlight actions
    def highlight(self, *args, section_index=None):
        """Issue highlight commands to a :class:`Code` block. See
        :func:`Code.highlight` for details on highlight specifiers.

        :param args: series of highlight specifiers
        :param section_indicator: The index number of the section to apply the
            highlighting to. Defaults to None, meaning use the last section.
            It is up to you to make sure the section is the kind that supports
            highlighting.
        """
        # Turn highlighting on and re-render
        if section_index is None:
            code = self.doc[-1]
        else:
            code = self.doc[section_index]

        code.highlight(*args)
        after = to_textual(self.doc)

        animate.cell_list.append(animate.Cell(self, after))
        return self

    def highlight_chain(self, *args, section_index=None):
        """Issue a series of highlight commands in sequence. Highlights the
        first, waits, turns it off, then highlights the next. Only works with
        :class:`Code` blocks. See :func:`Code.highlight` for details on
        highlight specifiers.

        :param args: series of highlight specifiers. Each argument can also
            be a list to activate a set of highlight specifiers together
        :param section_indicator: The index number of the section to apply the
            highlighting to. Defaults to None, meaning use the last section.
            It is up to you to make sure the section is the kind that supports
            highlighting.
        """
        # Turn highlighting on and re-render
        if section_index is None:
            code = self.doc[-1]
        else:
            code = self.doc[section_index]

        for arg in args:
            if isinstance(arg, list):
                specifiers = arg
            else:
                specifiers = [arg]

            code.highlight(*specifiers)
            after = to_textual(self.doc)
            animate.cell_list.append(animate.Cell(self, after))
            animate.cell_list.append(animate.WaitCell())

            code.highlight_off(*specifiers)
            after = to_textual(self.doc)
            animate.cell_list.append(animate.Cell(self, after))

        return self

    def highlight_off(self, *args, section_index=None):
        """Issue highlight_off commands to a :class:`Code` block. See
        :func:`Code.highlight_off` for details on highlight specifiers.

        :param args: series of highlight specifiers. Each argument can also
            be a list to deactivate a set of highlight specifiers together
        :param section_indicator: The index number of the section to apply the
            highlighting to. Defaults to None, meaning use the last section.
            It is up to you to make sure the section is the kind that supports
            highlighting.
        """
        # Turn highlighting off and re-render
        if section_index is None:
            code = self.doc[-1]
        else:
            code = self.doc[section_index]

        code.highlight_off(*args)
        after = to_textual(self.doc)

        animate.cell_list.append(animate.Cell(self, after))
        return self

    def highlight_all_off(self):
        """Removes all highlighting from all :class:`Code` containers"""
        # Turn all highlighting off and re-render
        for section in self.doc:
            if isinstance(section, Code):
                section.highlight_all_off()

        after = to_textual(self.doc)
        animate.cell_list.append(animate.Cell(self, after))
        return self



# scribe.py
# Print Utilities

from rich.console import Console

# ==============================================================================

console = Console(highlight=False)
rprint = console.print

# ==============================================================================

def print_code_lines(lines, title="", title_enabled=True, no_colour=False):
    if title_enabled:
        console.rule(title)

    for line in lines:
        rprint( (f"CodeLine({line.lexer_spec.lexer_cls.__name__}, "
            f"{line.has_newline}") )
        for part in line.parts:
            output = "   "
            if not no_colour:
                output += "[reverse]"

            output += part.text

            if not no_colour:
                output += "[/]"

            output += f"  -- {part.token}"
            rprint(output)

    if title_enabled:
        console.rule()



# content.py
#
# Encapsulates a block of code and manages any associated style information
import math

from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path

from purdy.parser import (CodeLine, CodePart, Fold, HighlightOff, HighlightOn,
    LexerSpec, LineNumber, Parser)
from purdy.themes import THEME_MAP, EMPTY_THEME

# ===========================================================================

class Section:
    """Abstract base class for classes that contain a list of lines which can
    be collected inside of a class:`Document`"""
    def render(self, render_state):
        for line_index, line in enumerate(self.lines):
            self.render_line(render_state, line, line_index)

    def render_line(self, render_state, line, line_index):
        raise NotImplementedError()

# ===========================================================================

@dataclass
class _CodeLineMetadata:
    ### Used to track data associated with a :class:`CodeLine` within a
    # :class:`Motif`.
    highlight: bool = False
    highlight_partial: list = field(default_factory=list)
    folded: bool = False
    hidden: bool = False


class Code(Section):
    """Encapsulates :class:`CodeLine` objects to track lines of code and any
    associated style information."""

    """Read code from a file, build an associated parser, and add the
    resulting lines to this object.

    :param filename: Name of file to read of :class:`pathlib.Path` object
    :param parser_identifier: Identifier for :class:`purdy.parser.Parser` to
        use when parsing the code. Defaults to "detect"
    :param theme: Either a string containing the base name of a theme (without
        the category type like code, con, etc) or a :class:Theme object.
        Defaults to `None` in which case it uses the class attribute
        `default_theme_name` as the theme name.
    """
    default_theme_name = "default"

    def __init__(self, filename, parser_identifier="detect", theme=None):
        # !!! If any defaults in here change make sure to update the .text()
        # factory
        lexer_spec = LexerSpec.get_spec(parser_identifier, hint=filename)
        self._pre_parse_init(theme, lexer_spec)

        path = Path(filename).resolve()
        self.parser = Parser(lexer_spec)
        self.parser.parse(path.read_text(), self)

        # !!! Anything added under here has to be copied to the text factory
        # and the spawn methods!!!

    @classmethod
    def text(cls, text, parser_identifier="py", theme=None):
        """Factory method for reading code from a string instead of a file.

        :param text: Text to parse
        :param parser_identifier: Identifier for :class:`purdy.parser.Parser` to
            use when parsing the code. Defaults to "py".
        """
        # A bit tricky: construct the object without invoking __init__
        obj = Code.__new__(Code)

        lexer_spec = LexerSpec.get_spec(parser_identifier)
        obj._pre_parse_init(theme, lexer_spec)

        obj.parser = Parser(lexer_spec)
        obj.parser.parse(text, obj)

        # !!! Anything added under here has to match __init__ and spawn!
        return obj

    def _pre_parse_init(self, theme, lexer_spec):
        ### Since the `.text` method does tricky stuff with `__new__`, common
        # initialization is done in this method
        self.lines = []
        self.current = 0

        self.meta = defaultdict(_CodeLineMetadata)

        if theme is None:
            theme = self.default_theme_name

        if isinstance(theme, str):
            self.theme = THEME_MAP[theme][lexer_spec.category]
        else:
            self.theme = theme

    def reset_metadata(self):
        """Sets all the style metadata back to defaults. Mostly used for
        testing.
        """
        self.meta = defaultdict(_CodeLineMetadata)

    # --- Accessors
    def __getitem__(self, index):
        """Returns a new `Code` object containing the subset of `CodeLine`
        objects indicated by `index`

        :param index: an integer or a slice
        :returns: A copy of this `Code` object but containing only a subset of
            ts lines
        """
        code = self.spawn()
        if isinstance(index, slice):
            code.lines = deepcopy(self.lines[index.start:index.stop])
        else:
            code.lines = [deepcopy(self.lines[index])]

        return code

    def spawn(self):
        """Returns a new :class:`Code` object with the same parser and theme
        as this one, but empty of any code lines."""
        obj = Code.__new__(Code)
        obj.lines = []
        obj.parser = self.parser
        obj.current = 0
        obj.meta = defaultdict(_CodeLineMetadata)
        obj.theme = self.theme

        return obj

    def chunk(self, amount):
        """This method is for doing iterator-like access to the `Code` object.
        Each call returns a copy of the object containing a subset of lines,
        with each subsequent call returning the next subset. To reset to the
        beginning, set :attr:`Code.current` to 0.

        :param amount: Number of `CodeLine` objects to  include in the
            result
        :returns: A copy of this `Code` object but containing only a subset of
            ts lines
        """
        code = self[self.current:self.current + amount]
        self.current += amount
        return code

    def remaining_chunk(self):
        """Associated with the :method:`Code.chunk` call, but returns whatever
        is left in the block.

        :returns: A copy of this `Code` object but containing the lines
            remaining in the chunkification process
        """
        code = self[self.current:]
        self.current = len(self.lines)
        return code

    # === Style Methods

    # --- Fold Management
    def fold(self, index, length):
        """Create a fold in the code

        :param index: Index number to start the fold at
        :param length: how many lines to include in the fold
        """
        # Nested folding is not supported, check for overlaps and error out
        for i in range(index, index + length):
            if i in self.meta and (self.meta[i].folded or self.meta[i].hidden):
                raise ValueError("Nested folding is not allowed")

        # Fold the line
        self.meta[index].folded = True
        for i in range(index + 1, index + length):
            self.meta[i].hidden = True

    def unfold(self, index):
        """Removes a previously created fold that starts on the given index
        line."""
        if index not in self.meta or (not self.meta[index].folded):
            raise ValueError(f"Line {index} was not the beginning of a fold")

        self.meta[index].folded = False

        # Loop through everything past the fold and unhide if hidden
        for i in range(index + 1, len(self.lines)):
            if i not in self.meta or not self.meta[i].hidden:
                break

            self.meta[i].hidden = False

    # --- Highlight Setters
    def _set_highlight(self, indicator, value):
        match indicator:
            case int(index):
                # Single number, highlight full line
                if index < 0:
                    index = len(self.lines) + index

                self.meta[index].highlight = value
            case (start, length):
                # Tuple, highlight start and length following lines
                if start < 0:
                    start = len(self.lines) + start

                for index in range(start, start + length):
                    self.meta[index].highlight = value
            case str(indicator):
                # String can be an integer, or a range
                indicator = indicator.strip()
                if indicator.startswith("-"):
                    # Negative Number
                    index = len(self.lines) + int(indicator)

                    self.meta[index].highlight = value
                elif "-" in indicator:
                    # Range
                    start, stop = indicator.split("-")

                    for index in range(int(start), int(stop) + 1):
                        self.meta[index].highlight = value
                else:
                    # Positive number
                    self.meta[int(indicator)].highlight = value

    def _parse_partial(self, indicator):
        indicator = indicator.replace(" ", "")
        index, scope = indicator.split(":")
        start, length = scope.split(",")

        # Meta storage is a dictionary so need to turn the negative index into
        # its positive equivalent as the dict doesn't know -1 is the last item
        # in the list
        index = int(index)
        if index < 0:
            # Negative indexing
            index = len(self.lines) + index

        return index, (int(start), int(length))

    def highlight(self, *args):
        """Turn highlighting on for one or more code lines. Each argument can
        be an int (index value, supports negative indexing), a tuple
        (specifying starting line and number of lines to highlight, start line
        can be negative), a string specifying a range ("1-3" highlights lines
        1 through 3 inclusive), or a partial highlight.

        Partial highlights support highlighting a subset of a line and are
        specified by a line number, a colon, a startng character position and
        a length.  Example "3:15,5" highlights characters 15-20 on index line
        3. The line number indicator supports negative indexing.

        All start positions are zero indexed.
        """
        for arg in args:
            if isinstance(arg, str) and ":" in arg:
                index, spec = self._parse_partial(arg)

                self.meta[index].highlight_partial.append(spec)
                continue

            # Argument is for a full line add it to the highlighting set
            self._set_highlight(arg, True)

    def highlight_off(self, *args):
        """Turns highlighting for one or more code lines. Each argument can be
        an int line index (negative indexing supported), a tuple (start
        indexing and number of lines to turn off) or a string specifying a range
        ("3-5" turns off line indexes 3 through 5).

        This turns off highlighting for both full lines and partial
        highlights. There is no way of turning off just a single partial
        inside of a line, if this is needed, turn off the line and turn
        partial back on for any segments you wish to keep.

        Turning off a line that is not highlighted is ignored.
        """
        for arg in args:
            if isinstance(arg, str) and ":" in arg:
                index, spec = self._parse_partial(arg)
                if index in self.meta:
                    # Only turn it off it was already on
                    try:
                        self.meta[index].highlight_partial.remove(spec)
                    except ValueError:
                        raise ValueError("Can only turn off existing partials")

                continue

            # Argument is for one or more full lines, turn them off
            self._set_highlight(arg, False)

    def highlight_all_off(self):
        """Removes all highlighting"""
        for value in self.meta.values():
            value.highlight = False
            value.highlight_partial = []

    def is_highlighted(self, line_index):
        if line_index not in self.meta:
            return False

        meta_info = self.meta[line_index]
        return meta_info.highlight or meta_info.highlight_partial

    # === Rendering
    def render_line(self, render_state, line, line_index):
        """Responsible for rendering the given line and appending the result
        into the :class:`RenderState` object.
        """
        if line_index in self.meta and self.meta[line_index].hidden:
            # No change to render_state, but may need to advance line count
            if render_state.doc.line_numbers_enabled:
                render_state.line_number += 1
            return

        if line_index in self.meta and self.meta[line_index].folded:
            # This is the parent line in a fold, display an indicator
            # instead
            output = CodeLine(lexer_spec=line.lexer_spec, has_newline=True)
            output.parts.append(CodePart(Fold, text=render_state.doc.fold_char))

            if render_state.doc.line_numbers_enabled:
                part = render_state.next_line_number_part()
                output.parts.insert(0, part)

            render_state.formatter.render_code_line(render_state, output)
            return

        # Handle the actual line, adding line numbers, dealing with
        # highlighting and wrapping it if needed
        if self.is_highlighted(line_index):
            line = self._apply_highlight(line_index)

        output = deepcopy(line)
        if render_state.doc.line_numbers_enabled:
            part = render_state.next_line_number_part()
            output.parts.insert(0, part)

        lwr = _LineWrapRenderer(render_state, output)
        lwr.run()

    # --- Highlighting Application
    @classmethod
    def _expand_parts(cls, line):
        """Creates a new list of parts expanding the text of each part into
        its own :class:`CodePart`, essentially creating parts consisting of a
        single letter.

        :param line: :class:`CodeLine` to expand
        :returns: list of :class:`CodePart` objects
        """
        char_parts = []
        for part in line.parts:
            if len(part.text) == 0:
                char_parts.append(deepcopy(part))
                continue

            for char in part.text:
                char_parts.append(CodePart(part.token, char))

        return char_parts

    @classmethod
    def _chop_partial_highlight(cls, line, cutpoints):
        """Creates a new CodeLine with highlight tokens at partial highlight
        spots

        :param line: CodeLine to copy and transform
        :param cutpoints: iterable of (start, length) cut point tuples
        """
        cutpoints.sort()
        char_parts = cls._expand_parts(line)

        # Loop through our single-character listing and insert the appropriate
        # highlight on and off tokens
        char_count = 0
        start = cutpoints[0][0]   # start char for first marker
        end = cutpoints[0][0] + cutpoints[0][1]  - 1 # end char for first marker
        cut_index = 0
        output = CodeLine(line.lexer_spec, has_newline=line.has_newline)
        highlighting = False
        for part in char_parts:
            if char_count == start and len(part.text) != 0 and not highlighting:
                output.parts.append(CodePart(HighlightOn, ""))
                output.parts.append(part)
                char_count += 1
                highlighting = True
                continue

            if char_count == end and len(part.text) != 0 and highlighting:
                char_count += 1
                output.parts.append(part)
                output.parts.append(CodePart(HighlightOff, ""))
                highlighting = False
                try:
                    # Was ended, advance to next cutpoint
                    cut_index += 1
                    start = cutpoints[cut_index][0]
                    end = cutpoints[cut_index][0] + cutpoints[0][1] - 1

                except IndexError:
                    # No cutpoints left, do nothing
                    pass

                continue

            output.parts.append(part)
            char_count += len(part.text)

        # Recompress our output
        output.compress()
        return output

    def _apply_highlight(self, line_index):
        """If there is highlighting to apply, creates a new :class:`CodeLine`
        as a copy of the given one but with highlight tokens applied. If there
        is no highlighting, just returns the given line.

        :param line_index: Index value of `CodeLine` inside this `Code` object
        """
        if line_index not in self.meta:
            # No highlighting (this is checked here to remove "and" clauses
            # from below reducing line length)
            return self.lines[line_index]

        output = deepcopy(self.lines[line_index])

        if self.meta[line_index].highlight:
            # Highlight whole line, stick tokens at beginning and end
            output.parts.insert(0, CodePart(HighlightOn, ""))
            output.parts.append(CodePart(HighlightOff, ""))
        elif self.meta[line_index].highlight_partial:
            # Partial highlighting, insert tokens as needed inside the line
            output = self._chop_partial_highlight(output,
                self.meta[line_index].highlight_partial)

        # Else: no highlighting
        return output

# ---------------------------------------------------------------------------

class _LineWrapRenderer:
    """Splits a line up into multiple parts based on the wrap length and
    updates renders content with each new line."""
    def __init__(self, render_state, line):
        self.render_state = render_state
        self.formatter = render_state.formatter
        self.line = line

    def run(self):
        # If wrapping is off
        if self.render_state.doc.wrap is None:
            self.formatter.render_code_line(self.render_state, self.line)
            return

        # If the line is shorter than the wrap
        line_width = self.line.parts.text_length
        if line_width < self.render_state.doc.wrap:
            self.formatter.render_code_line(self.render_state, self.line)
            return

        chunkify = deepcopy(self.line)
        self._chunk_line(chunkify)

    def _reset_current_line(self):
        self.current = self.line.spawn()
        self.length = 0

    def _split_part(self, part):
        # Split line here, -ve value to get chars from RHS
        cut_at = self.render_state.doc.wrap - self.length
        left_of = part.text[0:cut_at]
        split_point = left_of.rfind(" ")

        if split_point == -1:
            # No split point in the text, render what we have so far
            self.formatter.render_code_line(self.render_state,
                self.current)

            # Set up for next line
            self._reset_current_line()
            self.current.parts.append(part)
            self.length += len(part.text)
            return None

        # Stuff everything to the left of the split point into the
        # current line and render it
        left = CodePart(part.token, part.text[:split_point + 1])
        self.current.parts.append(left)
        self.formatter.render_code_line(self.render_state, self.current)

        # Everything else goes into the next line
        self._reset_current_line()
        right = CodePart(part.token, part.text[split_point + 1:])
        return right

    def _chunk_line(self, chunkify):
        """Splits given line into chunks of rendered lines. Modifies line in
        place, so it what is passed in should be a copy."""
        self._reset_current_line()

        for part in chunkify.parts:
            self.length += len(part.text)
            if self.length > self.render_state.doc.wrap:
                right = self._split_part(part)
                while(right):
                    if len(right.text) > self.render_state.doc.wrap:
                        right = self._split_part(right)
                    else:
                        self.length = len(right.text)
                        self.current.parts.append(right)
                        break
            else:
                # Not a split point, Copy the CodePart into the wrapped line
                full = CodePart(part.token, part.text)
                self.current.parts.append(full)

        if len(self.current.parts) > 0:
            # Render whatever is left
            self.formatter.render_code_line(self.render_state, self.current)

# ===========================================================================

class StringSection(Section):
    def __init__(self, content=None):
        super().__init__()
        self.theme = EMPTY_THEME

        if content is None:
            self.lines = []
        elif isinstance(content, str):
            self.lines = [content, ]
        elif isinstance(content, list):
            self.lines = content
        else:
            raise ValueError("StringSection requires string or list of strings")

    def render_line(self, render_state, line, line_index):
        # Strings have no formatting, just append it
        render_state.content += line

# ===========================================================================

class Document(list):
    """Container for renderable :class:`Section` objects like :class:`Code`
    and :class:`StringSection`. Constructor optionally takes a `section` to
    include. Manages display attributes that get used across multiple sections
    like line numbering.
    """
    def __init__(self, sections=None):
        super().__init__()

        if isinstance(sections, Section):
            self.append(sections)
        elif sections is not None:
            for item in sections:
                self.append(item)

        self.background = None
        self.line_numbers_enabled = False
        self.starting_line_number = 1
        self.line_number_gap_char = " "
        self.wrap = None
        self.fold_char = "⠇"

# ---------------------------------------------------------------------------

class RenderState:
    """Encapsulates the data needed for rendering and will contain the
    rendered output of of a :class:`Document`.

    :param document: `Document` that will be rendered
    :param future_length: When calculating width for line numbers, add this
        value to the max length. Useful when the `RenderState` needs to be
        created before content gets added to the `Document`.
    """
    def __init__(self, document, future_length=0):
        self.doc = document
        self.content = ""
        self.line_number = None

        if document.line_numbers_enabled:
            self.line_number = document.starting_line_number

            max_line = document.starting_line_number + future_length
            for section in document:
                max_line += len(section.lines)

            self.line_number_width = int(math.log10(max_line) + 1)

    def next_line_number(self):
        """Gets the next line number and increments the count."""
        output = self.line_number
        self.line_number += 1
        return f"{output:{self.line_number_width}d} "

    def next_line_number_part(self):
        """Gets the next line number as a :class:`CodePart` and increments the
        count."""
        output = self.line_number
        self.line_number += 1
        return CodePart(LineNumber, f"{output:{self.line_number_width}d} ")

    def line_number_gap(self):
        """Returns a whitespace string the width of a line number, useful for
        when dealing with wrapping
        """
        # Width of line number plus a gutter
        return (self.line_number_width + 1) * " "
